@@ -1,5 +1,4 @@
-/*
-    GNU LESSER GENERAL PUBLIC LICENSE
+/*    GNU LESSER GENERAL PUBLIC LICENSE
     Copyright (C) 2006 The Lobo Project
 
     This library is free software; you can redistribute it and/or
@@ -16,7 +15,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Sep 3, 2005
@@ -32,6 +31,7 @@ import org.lobobrowser.util.*;
 import org.lobobrowser.util.io.EmptyReader;
 import org.w3c.dom.*;
 import org.w3c.dom.css.CSSStyleSheet;
+import org.w3c.dom.views.*;
 import org.w3c.dom.html2.*;
 import org.xml.sax.ErrorHandler;
 import org.xml.sax.SAXException;
@@ -44,7 +44,10 @@ import java.util.*;
 import java.util.logging.*;
 import java.security.*;
 
-public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
+/**
+ * Implementation of the W3C <code>HTMLDocument</code> interface.
+ */
+public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument, DocumentView {
 	private static final Logger logger = Logger.getLogger(HTMLDocumentImpl.class.getName());
 	private final ElementFactory factory;
 	private final HtmlRendererContext rcontext;
@@ -150,6 +153,10 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	public void setDefaultTarget(String value) {
 		this.defaultTarget = value;
 	}
+
+	public AbstractView getDefaultView() {
+		return this.window;
+	}
 	
 	public String getTextContent() throws DOMException {
 		return null;
@@ -211,7 +218,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	public HTMLCollection getImages() {
 		synchronized(this) {
 			if(this.images == null) {
-				this.images = new FilteredHTMLCollectionImpl(this, this.elementsById, new ImageFilter(), this.treeLock);
+				this.images = new DescendentHTMLCollection(this, new ImageFilter(), this.treeLock);
 			}
 			return this.images;
 		}
@@ -221,7 +228,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 		synchronized(this) {
 			if(this.applets == null) {
 				//TODO: Should include OBJECTs that are applets?
-				this.applets = new FilteredHTMLCollectionImpl(this, this.elementsById, new AppletFilter(), this.treeLock);
+				this.applets = new DescendentHTMLCollection(this, new AppletFilter(), this.treeLock);
 			}
 			return this.applets;
 		}
@@ -230,7 +237,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	public HTMLCollection getLinks() {
 		synchronized(this) {
 			if(this.links == null) {
-				this.links = new FilteredHTMLCollectionImpl(this, this.elementsById, new LinkFilter(), this.treeLock);
+				this.links = new DescendentHTMLCollection(this, new LinkFilter(), this.treeLock);
 			}
 			return this.links;
 		}
@@ -239,7 +246,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	public HTMLCollection getForms() {
 		synchronized(this) {
 			if(this.forms == null) {
-				this.forms = new FilteredHTMLCollectionImpl(this, this.elementsById, new FormFilter(), this.treeLock);
+				this.forms = new DescendentHTMLCollection(this, new FormFilter(), this.treeLock);
 			}
 			return this.forms;
 		}
@@ -248,7 +255,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	public HTMLCollection getFrames() {
 		synchronized(this) {
 			if(this.frames == null) {
-				this.frames = new FilteredHTMLCollectionImpl(this, this.elementsById, new FrameFilter(), this.treeLock);
+				this.frames = new DescendentHTMLCollection(this, new FrameFilter(), this.treeLock);
 			}
 			return this.frames;
 		}
@@ -257,7 +264,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	public HTMLCollection getAnchors() {
 		synchronized(this) {
 			if(this.anchors == null) {
-				this.anchors = new FilteredHTMLCollectionImpl(this, this.elementsById, new AnchorFilter(), this.treeLock);
+				this.anchors = new DescendentHTMLCollection(this, new AnchorFilter(), this.treeLock);
 			}
 			return this.anchors;
 		}
@@ -273,6 +280,8 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 				// they should be able to get cookies on that document.
 				// Note that this Document instance cannot be created
 				// with an arbitrary URL.
+				
+				// TODO: Security: Review rationale.
 				public Object run() {
 					return ucontext.getCookie(documentURL);
 				}
@@ -501,7 +510,9 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 
 	public ProcessingInstruction createProcessingInstruction(
 			String target, String data) throws DOMException {
-		throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "HTML document");
+		HTMLProcessingInstruction node = new HTMLProcessingInstruction(target, data);
+		node.setOwnerDocument(this);
+		return node;
 	}
 
 	public Attr createAttribute(String name) throws DOMException {
@@ -761,7 +772,17 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 		}
 	}
 
-	private final Collection styleSheets = new LinkedList();
+	private final Collection styleSheets = new CSSStyleSheetList();
+	
+	public class CSSStyleSheetList extends ArrayList {
+		public int getLength(){
+			return this.size();
+		}
+		
+		public CSSStyleSheet item(int index){
+			return (CSSStyleSheet) get(index);
+		}
+	}
 	
 	final void addStyleSheet(CSSStyleSheet ss) {
 		synchronized(this.treeLock) {
@@ -785,6 +806,10 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 		this.allInvalidated();
 	}
 	
+	public Collection getStyleSheets(){
+		return this.styleSheets;
+	}
+	
 	private StyleSheetAggregator styleSheetAggregator = null;
 	
 	final StyleSheetAggregator getStyleSheetAggregator() {
@@ -802,9 +827,14 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 			return ssa;
 		}		
 	}
-
+	
 	private final ArrayList documentNotificationListeners = new ArrayList(1);
 	
+	/**
+	 * Adds a document notification listener, which is informed about
+	 * changes to the document.
+	 * @param listener An instance of {@link DocumentNotificationListener}.
+	 */
 	public void addDocumentNotificationListener(DocumentNotificationListener listener) {
 		ArrayList listenersList = this.documentNotificationListeners;
 		synchronized(listenersList) {
@@ -843,7 +873,8 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	 * Called if something such as a color or
 	 * decoration has changed. This would be 
 	 * something which does not affect the
-	 * rendered size.
+	 * rendered size, and can be revalidated
+	 * with a simple repaint.
 	 * @param node
 	 */
 	public void lookInvalidated(NodeImpl node) {
@@ -895,7 +926,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	/**
 	 * This is called when the node has changed, but
 	 * it is unclear if it's a size change or a look
-	 * change.
+	 * change. An attribute change should trigger this.
 	 * @param node
 	 */
 	public void invalidated(NodeImpl node) {
@@ -917,7 +948,32 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 			}
 		}
 	}
-	
+
+	/**
+	 * This is called when children of the node might
+	 * have changed.
+	 * @param node
+	 */
+	public void structureInvalidated(NodeImpl node) {
+		ArrayList listenersList = this.documentNotificationListeners;
+		int size;
+		synchronized(listenersList) {
+			size = listenersList.size();
+		}
+		// Traverse list outside synchronized block.
+		// (Shouldn't call listener methods in synchronized block.
+		// Deadlock is possible). But assume list could have
+		// been changed.
+		for(int i = 0; i < size; i++) {
+			try {
+				DocumentNotificationListener dnl = (DocumentNotificationListener) listenersList.get(i);
+				dnl.structureInvalidated(node);
+			} catch(IndexOutOfBoundsException iob) {
+				// ignore
+			}
+		}
+	}
+
 	public void nodeLoaded(NodeImpl node) {
 		ArrayList listenersList = this.documentNotificationListeners;
 		int size;
@@ -986,15 +1042,17 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 	private final ImageEvent BLANK_IMAGE_EVENT = new ImageEvent(this, null);
 	
 	/**
-	 * Loads images such that they are shared if from the same URI.
-	 * Informs listener immediately if an image is already known.
+	 * Loads images asynchronously such that they are shared if loaded 
+	 * simultaneously from the same URI.
+	 * Informs the listener immediately if an image is already known.
 	 * @param relativeUri
 	 * @param imageListener
 	 */
-	void loadImage(String relativeUri, ImageListener imageListener) {
+	protected void loadImage(String relativeUri, ImageListener imageListener) {
 		HtmlRendererContext rcontext = this.getHtmlRendererContext();
 		if(rcontext == null) {
 			// Ignore image loading when there's no renderer context.
+			// Consider Cobra users who are only using the parser.
 			imageListener.imageLoaded(BLANK_IMAGE_EVENT);
 			return;			
 		}
@@ -1027,35 +1085,48 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 				httpRequest.addReadyStateChangeListener(new ReadyStateChangeListener() {
 					public void readyStateChanged() {
 						if(httpRequest.getReadyState() == HttpRequest.STATE_COMPLETE) {
-							ImageEvent newEvent = new ImageEvent(HTMLDocumentImpl.this, httpRequest.getResponseImage());
+							java.awt.Image newImage = httpRequest.getResponseImage();
+							ImageEvent newEvent = newImage == null ? null : new ImageEvent(HTMLDocumentImpl.this, newImage);
 							ImageListener[] listeners;
 							synchronized(map) {
 								newInfo.imageEvent = newEvent;
 								newInfo.loaded = true;
-								listeners = newInfo.getListeners();
+								listeners = newEvent == null ? null : newInfo.getListeners();
 								// Must remove from map in the locked block
 								// that got the listeners. Otherwise a new
 								// listener might miss the event??
 								map.remove(urlText);
 							}
-							int llength = listeners.length;
-							for(int i = 0; i < llength; i++) {
-								// Call holding no locks
-								listeners[i].imageLoaded(newEvent);
+							if(listeners != null) {
+								int llength = listeners.length;
+								for(int i = 0; i < llength; i++) {
+									// Call holding no locks
+									listeners[i].imageLoaded(newEvent);
+								}
 							}
 						}
 					}
 				});
 				SecurityManager sm = System.getSecurityManager();
 				if(sm == null) {
-					httpRequest.open("GET", url, true);					
+					try {
+						httpRequest.open("GET", url, true);
+						httpRequest.send(null);
+					} catch(java.io.IOException thrown) {
+						logger.log(Level.WARNING, "loadImage()", thrown);
+					}
 				}
 				else {
 					AccessController.doPrivileged(new PrivilegedAction() {
 						public Object run() {
 							// Code might have restrictions on accessing
 							// items from elsewhere.
-							httpRequest.open("GET", url, true);
+							try {
+								httpRequest.open("GET", url, true);
+								httpRequest.send(null);
+							} catch(java.io.IOException thrown) {
+								logger.log(Level.WARNING, "loadImage()", thrown);
+							}
 							return null;
 						}
 					});
@@ -1123,8 +1194,7 @@ public class HTMLDocumentImpl extends NodeImpl implements HTMLDocument {
 
 	private class LinkFilter implements NodeFilter {
 		public boolean accept(Node node) {
-			String nodeName = node.getNodeName();
-			return "LINK".equalsIgnoreCase(nodeName);
+			return node instanceof HTMLLinkElement;
 		}
 	}
 
