@@ -15,7 +15,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Sep 3, 2005
@@ -49,22 +49,25 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		this.noStyleSheet = false;
 	}
 	
-	private volatile CSS2PropertiesImpl currentStyleDeclarationState;	
-	private volatile CSS2PropertiesImpl localStyleDeclarationState;	
+	private volatile AbstractCSS2Properties currentStyleDeclarationState;	
+	private volatile AbstractCSS2Properties localStyleDeclarationState;	
 
 	protected final void forgetLocalStyle() {
 		synchronized(this) {
 			this.currentStyleDeclarationState = null;
 			this.localStyleDeclarationState = null;
+			this.computedStyles = null;
 		}
 	}
 
 	protected final void forgetStyle(boolean deep) {
-		//TODO: Maybe this should redo style declarations in
-		//CSS, not the whole properties object?
+		//TODO: OPTIMIZATION: If we had a ComputedStyle map in
+		//window (Mozilla model) the map could be cleared in one shot.
 		synchronized(this) {
 			this.currentStyleDeclarationState = null;
+			this.computedStyles = null;
 			this.isHoverStyle = null;
+			this.hasHoverStyleByElement = null;
 			if(deep) {
 				java.util.ArrayList nl = this.nodeList;
 				if(nl != null) {
@@ -84,8 +87,8 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 	 * Gets the style object associated with the element.
 	 * It may return null only if the type of element does not handle stylesheets.
 	 */
-	public CSS2PropertiesImpl getCurrentStyle() {
-		CSS2PropertiesImpl sds;
+	public AbstractCSS2Properties getCurrentStyle() {
+		AbstractCSS2Properties sds;
 		synchronized(this) {
 			sds = this.currentStyleDeclarationState;
 			if(sds != null) {
@@ -94,11 +97,11 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		}
 		// Can't do the following in synchronized block (reverse locking order with document).
 		// First, add declarations from stylesheet
-		sds = this.addStyleSheetDeclarations(sds);
+		sds = this.addStyleSheetDeclarations(sds, this.getPseudoNames());
 		// Now add local style if any.
-		CSS2PropertiesImpl localStyle = this.getStyle();
+		AbstractCSS2Properties localStyle = this.getStyle();
 		if(sds == null) {
-			sds = new CSS2PropertiesImpl(this);
+			sds = new ComputedCSS2Properties(this);
 			sds.setLocalStyleProperties(localStyle);
 		}
 		else {
@@ -107,7 +110,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		synchronized(this) {
 			// Check if style properties were set while outside
 			// the synchronized block (can happen).
-			CSS2PropertiesImpl setProps = this.currentStyleDeclarationState;
+			AbstractCSS2Properties setProps = this.currentStyleDeclarationState;
 			if(setProps != null) {
 				return setProps;
 			}
@@ -116,58 +119,19 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		}
 	}
 	
-//	/**
-//	 * This method may return <code>null</code> if
-//	 * there's no style declaration applicable to
-//	 * the element.
-//	 */
-//	public CSS2PropertiesImpl getCurrentStyle() {
-//		if(this.noStyleSheet) {
-//			return null;
-//		}
-//		synchronized(this) {
-//			Object sdsObj = this.styleDeclarationNoCreate;
-//			if(sdsObj == INVALID_CSS) {
-//				CSS2PropertiesImpl sdsOrig = this.styleDeclarationState;
-//				if(sdsOrig != null) {
-//					this.styleDeclarationNoCreate = sdsOrig;
-//					return sdsOrig;
-//				}
-//			}
-//			else {
-//				return (CSS2PropertiesImpl) sdsObj;
-//			}
-//		}
-//		// Cannot run the following in synchronized block (reverse lock order with document).
-//		// First, add declarations from stylesheet
-//		CSS2PropertiesImpl sds = this.addStyleSheetDeclarations(null);
-//		// Now add local style if any
-//		CSS2PropertiesImpl localStyle = this.getStyleNoCreate();
-//		if(localStyle != null) {
-//			if(sds == null) {
-//				sds = new CSS2PropertiesImpl(this);
-//			}
-//			sds.setLocalStyleProperties(localStyle);
-//		}
-//		synchronized(this) {
-//			this.styleDeclarationNoCreate = sds;
-//		}
-//		return sds;
-//	}
-
 	/**
 	 * Gets the local style object associated with the element. The properties
 	 * object returned only includes properties from the local style attribute.
 	 * It may return null only if the type of element does not handle stylesheets.
 	 */
-	public CSS2PropertiesImpl getStyle() {
-		CSS2PropertiesImpl sds;
+	public AbstractCSS2Properties getStyle() {
+		AbstractCSS2Properties sds;
 		synchronized(this) {
 			sds = this.localStyleDeclarationState;
 			if(sds != null) {
 				return sds;
 			}
-			sds = new CSS2PropertiesImpl(this);
+			sds = new LocalCSS2Properties(this);
 			// Add any declarations in style attribute (last takes precedence).
 			String style = this.getAttribute("style");
 			if(style != null && style.length() != 0) {
@@ -179,7 +143,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 				} catch(Exception err) {
 					String id = this.getId();
 					String withId = id == null ? "" : " with ID '" + id + "'"; 
-					this.warn("Unable to parse style attribute value for element " + this.getTagName() + withId + ".", err);
+					this.warn("Unable to parse style attribute value for element " + this.getTagName() + withId + " in " + this.getDocumentURL() + ".", err);
 				}
 			}
 			this.localStyleDeclarationState = sds;
@@ -187,64 +151,55 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		// Synchronization note: Make sure getStyle() does not return multiple values.
 		return sds;
 	}
-	
-//	/**
-//	 * This method may return <code>null</code> if
-//	 * there's no style declaration applicable to
-//	 * the element.
-//	 */
-//	public CSS2PropertiesImpl getStyleNoCreate() {
-//		if(this.noStyleSheet) {
-//			return null;
-//		}
-//		synchronized(this) {
-//			Object sdsObj = this.localStyleDeclarationNoCreate;
-//			if(sdsObj == INVALID_CSS) {
-//				CSS2PropertiesImpl sdsOrig = this.localStyleDeclarationState;
-//				if(sdsOrig != null) {
-//					this.localStyleDeclarationNoCreate = sdsOrig;
-//					return sdsOrig;
-//				}
-//				// Otherwise, fall through.
-//			}
-//			else {
-//				return (CSS2PropertiesImpl) sdsObj;
-//			}
-//		}
-//		// Then add any declarations in style attribute (last takes precedence).
-//		CSS2PropertiesImpl sds = null;
-//		String style = this.getAttribute("style");
-//		if(style != null && style.length() != 0) {
-//			CSSOMParser parser = new CSSOMParser();
-//			InputSource inputSource = this.getCssInputSourceForDecl(style);
-//			try {
-//				CSSStyleDeclaration sd = parser.parseStyleDeclaration(inputSource);	
-//				if(sds == null) {
-//					sds = new CSS2PropertiesImpl(this);
-//				}
-//				sds.addStyleDeclaration(sd);
-//			} catch(Exception err) {
-//				String id = this.getId();
-//				String withId = id == null ? "" : " with ID '" + id + "'"; 
-//				this.warn("Unable to parse style attribute value for element " + this.getTagName() + withId + ".", err);
-//			}
-//		}
-//		synchronized(this) {
-//			this.localStyleDeclarationNoCreate = sds;
-//			if(sds != null) {
-//				Object gsp = this.styleDeclarationNoCreate;
-//				if(gsp == null && sds != null) {
-//					this.styleDeclarationNoCreate = INVALID_CSS;
-//					this.styleDeclarationState = null;
-//				}
-//				else if(gsp != INVALID_CSS) {
-//					((CSS2PropertiesImpl) gsp).setLocalStyleProperties(sds);
-//				}
-//			}
-//		}
-//		return sds;
-//	}
 
+	private Map computedStyles;
+	
+	public AbstractCSS2Properties getComputedStyle(String pseudoElement) {
+		if(pseudoElement == null) {
+			pseudoElement = "";
+		}
+		synchronized(this) {
+			Map cs = this.computedStyles;
+			if(cs != null) {
+				AbstractCSS2Properties sds = (AbstractCSS2Properties) cs.get(pseudoElement);
+				if(sds != null) {
+					return sds;
+				}
+			}
+		}
+		// Can't do the following in synchronized block (reverse locking order with document).
+		// First, add declarations from stylesheet
+		Set pes = pseudoElement.length() == 0 ? null : Collections.singleton(pseudoElement);
+		AbstractCSS2Properties sds = this.addStyleSheetDeclarations(null, pes);
+		// Now add local style if any.
+		AbstractCSS2Properties localStyle = this.getStyle();
+		if(sds == null) {
+			sds = new ComputedCSS2Properties(this);
+			sds.setLocalStyleProperties(localStyle);
+		}
+		else {
+			sds.setLocalStyleProperties(localStyle);
+		}		
+		synchronized(this) {
+			// Check if style properties were set while outside
+			// the synchronized block (can happen). We need to
+			// return instance already set for consistency.
+			Map cs = this.computedStyles;
+			if(cs == null) {
+				cs = new HashMap(2);
+				this.computedStyles = cs;
+			}
+			else {
+				AbstractCSS2Properties sds2 = (AbstractCSS2Properties) cs.get(pseudoElement);
+				if(sds2 != null) {
+					return sds2;
+				}				
+			}
+			cs.put(pseudoElement, sds);
+		}
+		return sds;
+	}
+	
 	public void setStyle(Object value) {
 		throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "Cannot set style property");
 	}	
@@ -319,7 +274,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
  * when the one passed is <code>null</code>.
  * @param style
  */	
-	protected final CSS2PropertiesImpl addStyleSheetDeclarations(CSS2PropertiesImpl style) {
+	protected final AbstractCSS2Properties addStyleSheetDeclarations(AbstractCSS2Properties style, Set pseudoNames) {
 		Node pn = this.parentNode;
 		if(pn == null) {
 			// do later
@@ -332,13 +287,13 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 			String[] classNameArray = Strings.split(classNames);
 			for(int i = classNameArray.length; --i >= 0;) {
 				String className = classNameArray[i];
-				Collection sds = this.findStyleDeclarations(elementName, id, className);
+				Collection sds = this.findStyleDeclarations(elementName, id, className, pseudoNames);
 				if(sds != null) {
 					Iterator sdsi = sds.iterator();
 					while(sdsi.hasNext()) {
 						CSSStyleDeclaration sd = (CSSStyleDeclaration) sdsi.next();
 						if(style == null) {
-							style = new CSS2PropertiesImpl(this);
+							style = new ComputedCSS2Properties(this);
 						}
 						style.addStyleDeclaration(sd);
 					}
@@ -348,13 +303,13 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		else {
 			String id = this.getId();
 			String elementName = this.getTagName();
-			Collection sds = this.findStyleDeclarations(elementName, id, null);
+			Collection sds = this.findStyleDeclarations(elementName, id, null, pseudoNames);
 			if(sds != null) {
 				Iterator sdsi = sds.iterator();
 				while(sdsi.hasNext()) {
 					CSSStyleDeclaration sd = (CSSStyleDeclaration) sdsi.next();
 					if(style == null) {
-						style = new CSS2PropertiesImpl(this);
+						style = new ComputedCSS2Properties(this);
 					}
 					style.addStyleDeclaration(sd);
 				}
@@ -366,14 +321,44 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 	private boolean isMouseOver = false;
 	
 	public void setMouseOver(boolean mouseOver) {
-		boolean prevMouseOver = this.isMouseOver;
-		this.isMouseOver = mouseOver;
-		if(this.hasHoverStyle() && prevMouseOver != mouseOver) {
-			this.informInvalid();
+		if(this.isMouseOver != mouseOver) {
+			// Change isMouseOver field before checking to invalidate.
+			this.isMouseOver = mouseOver;
+			// Check if descendents are affected (e.g. div:hover a { ... } )
+			this.invalidateDescendentsForHover();
+			if(this.hasHoverStyle()) {
+				//TODO: OPTIMIZATION: In some cases it should be much
+				//better to simply invalidate the "look" of the node.
+				this.informInvalid();				
+			}
 		}
 	}
 
+	private void invalidateDescendentsForHover() {
+		synchronized(this.treeLock) {
+			this.invalidateDescendentsForHoverImpl(this);
+		}
+	}
+	
+	private void invalidateDescendentsForHoverImpl(HTMLElementImpl ancestor) {
+		ArrayList nodeList = this.nodeList;
+		if(nodeList != null) {
+			int size = nodeList.size();
+			for(int i = 0; i < size; i++) {
+				Object node = nodeList.get(i);
+				if(node instanceof HTMLElementImpl) {
+					HTMLElementImpl descendent = (HTMLElementImpl) node;
+					if(descendent.hasHoverStyle(ancestor)) {
+						descendent.informInvalid();
+					}
+					descendent.invalidateDescendentsForHoverImpl(ancestor);
+				}
+			}
+		}
+	}
+	
 	private Boolean isHoverStyle = null;
+	private Map hasHoverStyleByElement = null;
 	
 	private boolean hasHoverStyle() {
 		Boolean ihs;
@@ -396,7 +381,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 			if(classNames != null && classNames.length() != 0) {
 				classNameArray = Strings.split(classNames);
 			}
-			ihs = Boolean.valueOf(ssa.hasPseudoName(this, elementName, id, classNameArray, "hover"));
+			ihs = Boolean.valueOf(ssa.affectedByPseudoNameInAncestor(this, this, elementName, id, classNameArray, "hover"));
 		}
 		synchronized(this) {
 			this.isHoverStyle = ihs;
@@ -404,7 +389,51 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		return ihs.booleanValue();
 	}
 
-	private Set getPseudoNames() {
+	private boolean hasHoverStyle(HTMLElementImpl ancestor) {
+		Map ihs;
+		synchronized(this) {
+			ihs = this.hasHoverStyleByElement;
+			if(ihs != null) {
+				Boolean f = (Boolean) ihs.get(ancestor);
+				if(f != null) {
+					return f.booleanValue();
+				}
+			}
+		}
+		Boolean hhs;
+		HTMLDocumentImpl doc = (HTMLDocumentImpl) this.document;
+		if(doc == null) {
+			hhs = Boolean.FALSE;
+		}
+		else {
+			StyleSheetAggregator ssa = doc.getStyleSheetAggregator();
+			String id = this.getId();
+			String elementName = this.getTagName();
+			String classNames = this.getClassName();
+			String[] classNameArray = null;
+			if(classNames != null && classNames.length() != 0) {
+				classNameArray = Strings.split(classNames);
+			}
+			hhs = Boolean.valueOf(ssa.affectedByPseudoNameInAncestor(this, ancestor, elementName, id, classNameArray, "hover"));
+		}
+		synchronized(this) {
+			ihs = this.hasHoverStyleByElement;
+			if(ihs == null) {
+				ihs = new HashMap(2);
+				this.hasHoverStyleByElement = ihs;
+			}			
+			ihs.put(ancestor, hhs);
+		}
+		return hhs.booleanValue();
+	}
+
+	/**
+	 * Gets the pseudo-element lowercase names currently
+	 * applicable to this element. Method must return
+	 * <code>null</code> if there are no such
+	 * pseudo-elements.
+	 */
+	public Set getPseudoNames() {
 		Set pnset = null;
 		if(this.isMouseOver) {
 			if(pnset == null) {
@@ -415,13 +444,13 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		return pnset;
 	}
 	
-	protected final Collection findStyleDeclarations(String elementName, String id, String className) {
+	protected final Collection findStyleDeclarations(String elementName, String id, String className, Set pseudoNames) {
 		HTMLDocumentImpl doc = (HTMLDocumentImpl) this.document;
 		if(doc == null) {
 			return null;
 		}
 		StyleSheetAggregator ssa = doc.getStyleSheetAggregator();
-		return ssa.getStyleDeclarations(this, elementName, id, className, this.getPseudoNames());
+		return ssa.getStyleDeclarations(this, elementName, id, className, pseudoNames);
 	}
 	
 	public void informInvalid() {
@@ -568,14 +597,6 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		}
 	}
 
-	public String getInnerHTML() {
-		StringBuffer buffer = new StringBuffer();
-		synchronized(this) {
-			this.appendInnerHTMLImpl(buffer);
-		}
-		return buffer.toString();
-	}
-
 	public String getOuterHTML() {
 		StringBuffer buffer = new StringBuffer();
 		synchronized(this) {
@@ -584,27 +605,6 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		return buffer.toString();
 	}
 
-	protected void appendInnerHTMLImpl(StringBuffer buffer) {
-		ArrayList nl = this.nodeList;
-		int size;
-		if (nl != null && (size = nl.size()) > 0) {
-			for (int i = 0; i < size; i++) {
-				Node child = (Node) nl.get(i);
-				if (child instanceof HTMLElementImpl) {
-					((HTMLElementImpl) child).appendOuterHTMLImpl(buffer);
-				}
-				else if(child instanceof Comment) {
-					buffer.append("<!--" + ((Comment) child).getTextContent() + "-->");
-				}
-				else if (child instanceof Text) {
-					String text = ((Text) child).getTextContent();
-					String encText = Strings.strictHtmlEncode(text);
-					buffer.append(encText);
-				}
-			}
-		}
-	}
-
 	protected void appendOuterHTMLImpl(StringBuffer buffer) {
 		String tagName = this.getTagName();
 		buffer.append('<');
@@ -643,7 +643,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 	}
 	
 	public int getOffsetTop() {
-		CSS2PropertiesImpl style = this.getCurrentStyle();
+		AbstractCSS2Properties style = this.getCurrentStyle();
 		if(style == null) {
 			return 0;
 		}
@@ -652,7 +652,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 	}
 	
 	public int getOffsetLeft() {
-		CSS2PropertiesImpl style = this.getCurrentStyle();
+		AbstractCSS2Properties style = this.getCurrentStyle();
 		if(style == null) {
 			return 0;
 		}
@@ -661,7 +661,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 	}
 
 	public int getOffsetWidth() {
-		CSS2PropertiesImpl style = this.getCurrentStyle();
+		AbstractCSS2Properties style = this.getCurrentStyle();
 		if(style == null) {
 			return 0;
 		}
@@ -670,7 +670,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 	}
 
 	public int getOffsetHeight() {
-		CSS2PropertiesImpl style = this.getCurrentStyle();
+		AbstractCSS2Properties style = this.getCurrentStyle();
 		if(style == null) {
 			return 0;
 		}
@@ -678,7 +678,7 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		return valueText == null ? 0 : HtmlValues.getPixelSize(valueText, this.getRenderState(), 0);
 	}
 
-	public CSS2PropertiesImpl getParentStyle() {
+	public AbstractCSS2Properties getParentStyle() {
 		Object parent = this.parentNode;
 		if(parent instanceof HTMLElementImpl) {
 			return ((HTMLElementImpl) parent).getCurrentStyle();
@@ -686,6 +686,16 @@ public class HTMLElementImpl extends ElementImpl implements HTMLElement, CSS2Pro
 		return null;
 	}
 	
+	public String getDocumentBaseURI() {
+		HTMLDocumentImpl doc = (HTMLDocumentImpl) this.document;
+		if(doc != null) {
+			return doc.getBaseURI();
+		}
+		else {
+			return null;
+		}
+	}
+
 	public String toString() {
 		return super.toString() + "[currentStyle=" + this.getCurrentStyle() + "]";
 	}
