@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Sep 3, 2005
@@ -65,7 +65,8 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 	
 	/**
 	 * Tries to get a UINode associated with the current node.
-	 * Failing that, it tries ancestors recursively.
+	 * Failing that, it tries ancestors recursively. This method
+	 * will return the closest <i>block-level</i> renderer node, if any.
 	 */
 	public UINode findUINode() {
 		// Called in GUI thread always.
@@ -77,12 +78,6 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 		return parentNode == null ? null : parentNode.findUINode();
 	}
 		
-//	protected final Object getTreeLock() {
-//		//TODO: Is this necessary? Why not use a lock per node?
-//		Object doc = this.document;
-//		return doc == null ? this : doc;
-//	}
-//	
 	public Node appendChild(Node newChild)
 			throws DOMException {
 		synchronized(this.treeLock) {
@@ -98,7 +93,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 		}
 				
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 		return newChild;
 	}
@@ -118,7 +113,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}		
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 	}
 
@@ -158,13 +153,13 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 	}
 	
 	/**
-	 * Gets descendent nodes that match according to
-	 * the filter, but it does not nest into matching nodes.
+	 * Creates an <code>ArrayList</code> of descendent nodes that 
+	 * the given filter condition.
 	 */
-	public ArrayList getDescendents(NodeFilter filter) {
+	public ArrayList getDescendents(NodeFilter filter, boolean nestIntoMatchingNodes) {
 		ArrayList al = new ArrayList();
 		synchronized(this.treeLock) {
-			this.extractDescendentsArrayImpl(filter, al);
+			this.extractDescendentsArrayImpl(filter, al, nestIntoMatchingNodes);
 		}
 		return al;
 	}
@@ -175,7 +170,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 	 * @param filter
 	 * @param al
 	 */
-	private void extractDescendentsArrayImpl(NodeFilter filter, ArrayList al) {
+	private void extractDescendentsArrayImpl(NodeFilter filter, ArrayList al, boolean nestIntoMatchingNodes) {
 		ArrayList nl = this.nodeList;
 		if(nl != null) {
 			Iterator i = nl.iterator();
@@ -183,9 +178,12 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 				NodeImpl n = (NodeImpl) i.next();
 				if(filter.accept(n)) {
 					al.add(n);
+					if(nestIntoMatchingNodes) {
+						n.extractDescendentsArrayImpl(filter, al, nestIntoMatchingNodes);						
+					}
 				}
 				else if(n.getNodeType() == Node.ELEMENT_NODE) {
-					n.extractDescendentsArrayImpl(filter, al);
+					n.extractDescendentsArrayImpl(filter, al, nestIntoMatchingNodes);
 				}
 			}
 		}
@@ -378,7 +376,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 		return newChild;
 	}
@@ -397,7 +395,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 		return newChild;
 	}
@@ -413,7 +411,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			nl.set(idx, newChild);
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 		return newChild;
 	}
@@ -427,7 +425,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 		return oldChild;
 	}
@@ -448,7 +446,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		} finally {
 			if(!this.notificationsSuspended) {
-				this.informInvalid();
+				this.informStructureInvalid();
 			}
 		}
 	}
@@ -651,7 +649,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 	}
 	
@@ -660,7 +658,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			this.removeChildrenImpl(filter);
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 	}
 	
@@ -690,7 +688,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 		return newChild;
 	}
@@ -731,7 +729,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}		
 		} finally {
 			if(!this.notificationsSuspended) {
-				this.informInvalid();
+				this.informStructureInvalid();
 			}
 		}
 	}
@@ -775,7 +773,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}		
 		} finally {
 			if(!this.notificationsSuspended) {
-				this.informInvalid();
+				this.informStructureInvalid();
 			}
 		}
 	}
@@ -848,7 +846,7 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			}
 		}		
 		if(!this.notificationsSuspended) {
-			this.informInvalid();
+			this.informStructureInvalid();
 		}
 	}
 	
@@ -998,7 +996,16 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 			doc.invalidated(this);
 		}
 	}
-	
+
+	public void informStructureInvalid() {
+		// This is called when an attribute or child changes.
+		this.forgetRenderState();
+		HTMLDocumentImpl doc = (HTMLDocumentImpl) this.document;
+		if(doc != null) {
+			doc.structureInvalidated(this);
+		}
+	}
+
 	protected void informNodeLoaded() {
 		// This is called when an attribute or child changes.
 		this.forgetRenderState();
@@ -1082,4 +1089,35 @@ public abstract class NodeImpl extends AbstractScriptableDelegate implements Nod
 		}
 	}
 
+	public String getInnerHTML() {
+		StringBuffer buffer = new StringBuffer();
+		synchronized(this) {
+			this.appendInnerHTMLImpl(buffer);
+		}
+		return buffer.toString();
+	}
+	
+	protected void appendInnerHTMLImpl(StringBuffer buffer) {
+		ArrayList nl = this.nodeList;
+		int size;
+		if (nl != null && (size = nl.size()) > 0) {
+			for (int i = 0; i < size; i++) {
+				Node child = (Node) nl.get(i);
+				if (child instanceof HTMLElementImpl) {
+					((HTMLElementImpl) child).appendOuterHTMLImpl(buffer);
+				}
+				else if(child instanceof Comment) {
+					buffer.append("<!--" + ((Comment) child).getTextContent() + "-->");
+				}
+				else if (child instanceof Text) {
+					String text = ((Text) child).getTextContent();
+					String encText = Strings.strictHtmlEncode(text);
+					buffer.append(encText);
+				}
+				else if (child instanceof ProcessingInstruction) {
+					buffer.append(child.toString());
+				}
+			}
+		}
+	}
 }
