@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Apr 16, 2005
@@ -41,7 +41,10 @@ import java.util.*;
 /**
  * A Swing component that renders a HTML block, given 
  * by a DOM root or an internal element, typically a DIV.
- * This component cannot render FRAMESETs.
+ * This component <i>cannot</i> render FRAMESETs, which is done
+ * by {@link FrameSetPanel}. <code>HtmlBlockPanel</code>
+ * is used by {@link HtmlPanel} whenever the DOM
+ * is determined <i>not</i> to be a FRAMESET.
  * @see HtmlPanel
  * @see FrameSetPanel
  * @author J. H. S.
@@ -58,11 +61,11 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 	protected RBlock rblock;
 	protected int preferredWidth = -1;
 
-	public HtmlBlockPanel(int listNesting, UserAgentContext pcontext, HtmlRendererContext rcontext, FrameContext frameContext) {
-		this(listNesting, ColorFactory.TRANSPARENT, false, pcontext, rcontext, frameContext);
+	public HtmlBlockPanel(UserAgentContext pcontext, HtmlRendererContext rcontext, FrameContext frameContext) {
+		this(ColorFactory.TRANSPARENT, false, pcontext, rcontext, frameContext);
 	}
 	
-	public HtmlBlockPanel(int listNesting, Color background, boolean opaque, UserAgentContext pcontext, HtmlRendererContext rcontext, FrameContext frameContext) {
+	public HtmlBlockPanel(Color background, boolean opaque, UserAgentContext pcontext, HtmlRendererContext rcontext, FrameContext frameContext) {
 		this.setLayout(null);
 		this.setAutoscrolls(true);
 		this.frameContext = frameContext;
@@ -78,8 +81,10 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 				}
 			}
 		};
-		this.registerKeyboardAction(actionListener, "copy", KeyStroke.getKeyStroke(KeyEvent.VK_COPY, 0), JComponent.WHEN_FOCUSED);		
-		this.registerKeyboardAction(actionListener, "copy", KeyStroke.getKeyStroke(KeyEvent.VK_C, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()), JComponent.WHEN_FOCUSED);		
+		if(!GraphicsEnvironment.isHeadless()) {
+			this.registerKeyboardAction(actionListener, "copy", KeyStroke.getKeyStroke(KeyEvent.VK_COPY, 0), JComponent.WHEN_FOCUSED);		
+			this.registerKeyboardAction(actionListener, "copy", KeyStroke.getKeyStroke(KeyEvent.VK_C, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()), JComponent.WHEN_FOCUSED);
+		}
 		this.addMouseListener(new MouseListener() {
 			public void mouseClicked(MouseEvent e) {
 				onMouseClick(e);
@@ -118,6 +123,131 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 		});
 	}
 
+	/**
+	 * Scrolls the body area to the given location.
+	 * <p>
+	 * This method should be called from the GUI thread.
+	 * @param bounds The bounds in the scrollable block area that should
+	 *               become visible.
+	 * @param xIfNeeded If this parameter is true, scrolling will only occur if the
+	 *                  requested bounds are not currently visible horizontally.
+	 * @param yIfNeeded If this parameter is true, scrolling will only occur if the
+	 *                  requested bounds are not currently visible vertically.
+	 */
+	public void scrollTo(Rectangle bounds, boolean xIfNeeded, boolean yIfNeeded) {
+		RBlock block = this.rblock;
+		if(block != null) {
+			block.scrollTo(bounds, xIfNeeded, yIfNeeded);
+		}
+	}
+
+	/**
+	 * Scrolls the body area to the node given, if it is
+	 * part of the current document.
+	 * <p>
+	 * This method should be called from the GUI thread.
+	 * @param node A DOM node.
+	 */
+	public void scrollTo(Node node) {
+		Rectangle bounds = this.getNodeBounds(node, true);
+		if(bounds == null) {
+			return;
+		}
+		this.scrollTo(bounds, true, false);
+	}
+
+	/**
+	 * Gets the rectangular bounds of the given node.
+	 * <p>
+	 * This method should be called from the GUI thread.
+	 * @param node A node in the current document.
+	 * @param relativeToScrollable Whether the bounds should be relative to the
+	 *                             scrollable body area. Otherwise, they are
+	 *                             relative to the root block (which is the
+	 *                             essentially the same as being relative to this 
+	 *                             <code>HtmlBlockPanel</code> minus Swing
+	 *                             borders).
+	 */
+	public Rectangle getNodeBounds(Node node, boolean relativeToScrollable) {
+		RBlock block = this.rblock;
+		if(block == null) {
+			return null;
+		}
+		// Find UINode first
+		Node currentNode = node;
+		UINode uiNode = null;
+		while(currentNode != null) {
+			if(currentNode instanceof HTMLElementImpl) {
+				HTMLElementImpl element = (HTMLElementImpl) currentNode;
+				uiNode = element.getUINode();
+				if(uiNode != null) {
+					break;
+				}
+			}
+			currentNode = currentNode.getParentNode();
+		}
+		if(uiNode == null) {
+			return null;
+		}
+		RCollection relativeTo = relativeToScrollable ? (RCollection) block.getRBlockViewport() : (RCollection) block;
+		if(node == currentNode) {
+			BoundableRenderable br = (BoundableRenderable) uiNode;
+			Point guiPoint = br.getOriginRelativeTo(relativeTo);
+			Dimension size = br.getSize();
+			return new Rectangle(guiPoint, size);
+		}
+		else {
+			return this.scanNodeBounds((RCollection) uiNode, node, relativeTo);
+		}
+	}
+
+	/**
+	 * Gets an aggregate of the bounds of renderer leaf nodes.
+	 */
+	private Rectangle scanNodeBounds(RCollection root, Node node, RCollection relativeTo) {
+		Iterator i = root.getRenderables();
+		Rectangle resultBounds = null;
+		BoundableRenderable prevBoundable = null;
+		if(i != null) {
+			while(i.hasNext()) {
+				Renderable r = (Renderable) i.next();
+				Rectangle subBounds = null;
+				if(r instanceof RCollection) {
+					RCollection rc = (RCollection) r;
+					prevBoundable = rc;
+					subBounds = this.scanNodeBounds(rc, node, relativeTo);
+				}
+				else if(r instanceof BoundableRenderable) {
+					BoundableRenderable br = (BoundableRenderable) r;
+					prevBoundable = br;
+					if(Nodes.isSameOrAncestorOf(node, (Node) r.getModelNode())) {
+						Point origin = br.getOriginRelativeTo(relativeTo);
+						Dimension size = br.getSize();
+						subBounds = new Rectangle(origin, size);
+					}	
+				}
+				else {
+					// This would have to be a RStyleChanger. We rely on these
+					// when the target node has blank content.
+					if(Nodes.isSameOrAncestorOf(node, (Node) r.getModelNode())) {
+						int xInRoot = prevBoundable == null ? 0 : prevBoundable.getX() + prevBoundable.getWidth();
+						Point rootOrigin = root.getOriginRelativeTo(relativeTo);
+						subBounds = new Rectangle(rootOrigin.x + xInRoot, rootOrigin.y, 0, root.getHeight());
+					}
+				}
+				if(subBounds != null) {
+					if(resultBounds == null) {
+						resultBounds = subBounds;
+					}
+					else {
+						resultBounds = subBounds.union(resultBounds);
+					}
+				}
+			}
+		}
+		return resultBounds;
+	}
+	
 	public BoundableRenderable getRootRenderable() {
 		return this.rblock;
 	}
@@ -151,7 +281,7 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 			if(block != null) {
 				// Potentially does layout outside of GUI thread.
 				// Do not provide default overflow.
-				block.layout(pw, 0, false, false, RBlock.OVERFLOW_NONE);
+				block.layout(pw, 0, RBlock.OVERFLOW_NONE);
 				// Adjust for permanent vertical scrollbar.
 				int newPw = Math.max(block.width + block.getVScrollBarWidth(), pw);
 				return new Dimension(newPw, block.height);
@@ -176,7 +306,7 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 		}
 	}
 	
-	private Insets defaultPaddingInsets = null;
+	protected Insets defaultPaddingInsets = null;
 
 	public void setDefaultPaddingInsets(Insets insets) {
 		this.defaultPaddingInsets = insets;
@@ -243,13 +373,6 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 		toValidate.validate();
 	}
 
-	/**
-	 * @deprecated setContainer is unused.
-	 */
-	public void setRootNode(NodeImpl node, boolean setContainer) {
-		this.setRootNode(node);
-	}
-
 	protected void revalidatePanel() {
 		// Called in the GUI thread.
 		this.invalidate();
@@ -299,25 +422,6 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 			else {
 				this.frameContext.resetSelection(null);
 			}
-			
-//			BoundableRenderable r = block.getRenderable(point);
-//			if(r != null) {
-//				this.mousePressTarget = r;
-//				Point rpoint = r.getRenderablePoint(point.x, point.y);
-//				int rx = rpoint.x;
-//				int ry = rpoint.y;
-//				r.onMousePressed(event, rx, ry);
-//				RenderableSpot rp = r.getLowestRenderableSpot(rx, ry);
-//				if(rp != null) {
-//					this.frameContext.resetSelection(rp);
-//				}
-//				else {
-//					this.frameContext.resetSelection(null);
-//				}
-//			}
-//			else {
-//				this.frameContext.resetSelection(null);
-//			}
 		}
 	}
 	
@@ -332,21 +436,6 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 				block.onMouseClick(event, rx, ry);
 			}
 			block.onMouseReleased(event, rx, ry);
-			
-//			BoundableRenderable r = block.getRenderable(point);
-//			if(r != null) {
-//				Point rpoint = r.getRenderablePoint(point.x, point.y);
-//				int rx = rpoint.x;
-//				int ry = rpoint.y;
-//				// First, generate a mouse-click event if it's BUTTON1.
-//				if(event.getButton() == MouseEvent.BUTTON1) {
-//					// TODO: This will be raised twice on a double-click.
-//					r.onMouseClick(event, rx, ry);
-//				}
-//				// Now generate mouse-released event.
-//				r.onMouseReleased(event, rx, ry);
-//			}
-			
 			BoundableRenderable oldTarget = this.mousePressTarget;
 			if(oldTarget != null) {
 				this.mousePressTarget = null;
@@ -444,35 +533,37 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 		
 	
 	public void doLayout() {
-		if(EventQueue.isDispatchThread()) {
+		try {
 			Dimension size = this.getSize();
 			boolean liflag = loggableInfo;
 			long time1 = 0;
 			if(liflag) {
 				time1 = System.currentTimeMillis();
 			}
-			this.removeAll();
+			this.clearComponents();
 			RBlock block = this.rblock;
 			if(block != null) {
 				ModelNode rootNode = block.getModelNode();
-				block.layout(size.width, size.height, true, true);
+				block.layout(size.width, size.height);
+				// Must also adjust here
+				block.adjust(size.width, size.height, true, true, null);
 				//Only set origin
-				//this.rblock.setBounds(0, 0, size.width, size.height);
 				block.setOrigin(0, 0);
 				block.updateWidgetBounds(0, 0);
+				this.updateGUIComponents();
 				if(liflag) {
 					long time2 = System.currentTimeMillis();
 					String uri = rootNode instanceof Document ? ((Document) rootNode).getDocumentURI() : "";
 					logger.info("doLayout(): URI=[" + uri + "]. Block layout elapsed: " + (time2 - time1) + " ms. Component count: " + this.getComponentCount() + ".");
+				}					
+			}
+			else {
+				if(this.getComponentCount() > 0) {
+					this.removeAll();
 				}
 			}
-		}
-		else {
-			EventQueue.invokeLater(new Runnable() {
-				public void run() {
-					HtmlBlockPanel.this.doLayout();
-				}
-			});
+		} catch(Throwable thrown) {
+			logger.log(Level.SEVERE, "Unexpected error in layout engine. Document is " + this.getRootNode(), thrown);
 		}
 	}
 	
@@ -665,4 +756,61 @@ public class HtmlBlockPanel extends JComponent implements NodeRenderer, Renderab
 
 	public void clearDelayedPairs() {
 	}
+
+	private Set components;
+
+	private void clearComponents() {
+		Set c = this.components;
+		if(c != null) {
+			c.clear();
+		}
+	}
+	
+	public Component addComponent(Component component) {
+		Set c = this.components;
+		if(c == null) {
+			c = new HashSet();
+			this.components = c;
+		}
+		if(c.add(component)) {
+			return component;
+		}
+		else {
+			return null;
+		}
+	}
+	
+	private void updateGUIComponents() {
+		// We use this method, instead of removing all components and
+		// adding them back, because removal of components can cause
+		// them to lose focus.
+		
+		Set c = this.components;
+		if(c == null && this.getComponentCount() != 0) {
+			this.removeAll();
+		}
+		else {
+			// Remove children not in the set.
+			Set workingSet = new HashSet();
+			workingSet.addAll(c);
+			int count = this.getComponentCount();
+			for(int i = 0; i < count;) {
+				Component component = this.getComponent(i);
+				if(!c.contains(component)) {
+					this.remove(i);
+					count = this.getComponentCount();
+				}
+				else {
+					i++;
+					workingSet.remove(component);
+				}
+			}
+			// Add components in set that were not previously children.
+			Iterator wsi = workingSet.iterator();
+			while(wsi.hasNext()) {
+				Component component = (Component) wsi.next();
+				this.add(component);
+			}
+		}
+	}
 }
\ No newline at end of file
