@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Nov 19, 2005
@@ -41,7 +41,10 @@ import org.w3c.dom.html2.*;
 
 /**
  * The <code>HtmlPanel</code> class is a Swing
- * component that can render a HTML DOM. 
+ * component that can render a HTML DOM. It uses
+ * either {@link HtmlBlockPanel} or {@link FrameSetPanel}
+ * internally, depending on whether the document
+ * is determined to be a FRAMESET or not.
  * @author J. H. S.
  */
 public class HtmlPanel extends JComponent implements FrameContext {
@@ -78,8 +81,8 @@ public class HtmlPanel extends JComponent implements FrameContext {
 	 * This method currently does not have any effect when the
 	 * document is a FRAMESET.
 	 * <p>
-	 * Note also that setting the preferred width to a value other
-	 * than <code>-1</code> (the default) will have an effect in renderer performance.	 
+	 * Note also that setting the preferred width (to a value other
+	 * than <code>-1</code>) will negatively impact performance.	 
 	 *  
 	 * @param width The preferred width, or <code>-1</code> to unset.
 	 */
@@ -91,6 +94,39 @@ public class HtmlPanel extends JComponent implements FrameContext {
 		}
 	}
 	
+	/**
+	 * If the current document is not a FRAMESET,
+	 * this method scrolls the body area to the given location.
+	 * <p>
+	 * This method should be called from the GUI thread.
+	 * @param bounds The bounds in the scrollable block area that should
+	 *               become visible.
+	 * @param xIfNeeded If this parameter is true, scrolling will only occur if the
+	 *                  requested bounds are not currently visible horizontally.
+	 * @param yIfNeeded If this parameter is true, scrolling will only occur if the
+	 *                  requested bounds are not currently visible vertically.
+	 */
+	public void scrollTo(Rectangle bounds, boolean xIfNeeded, boolean yIfNeeded) {
+		HtmlBlockPanel htmlBlock = this.htmlBlock;
+		if(htmlBlock != null) {
+			htmlBlock.scrollTo(bounds, xIfNeeded, yIfNeeded);
+		}		
+	}
+
+	/**
+	 * Scrolls the body area to the node given, if it is
+	 * part of the current document.
+	 * <p>
+	 * This method should be called from the GUI thread.
+	 * @param node A DOM node.
+	 */
+	public void scrollTo(org.w3c.dom.Node node) {
+		HtmlBlockPanel htmlBlock = this.htmlBlock;
+		if(htmlBlock != null) {
+			htmlBlock.scrollTo(node);
+		}				
+	}
+	
 	/**
 	 * Gets the root <code>Renderable</code> of
 	 * the HTML block. It returns <code>null</code>
@@ -148,7 +184,7 @@ public class HtmlPanel extends JComponent implements FrameContext {
 	 * It is made available so it can be overridden.
 	 */
 	protected HtmlBlockPanel createHtmlBlockPanel(UserAgentContext ucontext, HtmlRendererContext rcontext) {
-		return new HtmlBlockPanel(0, java.awt.Color.WHITE, true, ucontext, rcontext, this);		
+		return new HtmlBlockPanel(java.awt.Color.WHITE, true, ucontext, rcontext, this);
 	}
 
 	/**
@@ -162,13 +198,28 @@ public class HtmlPanel extends JComponent implements FrameContext {
 	/**
 	 * Scrolls the document such that x and y coordinates
 	 * are placed in the upper-left corner of the panel.
+	 * <p>
+	 * This method may be called outside of the GUI Thread.
 	 * @param x The x coordinate.
 	 * @param y The y coordinate.
 	 */
-	public void scroll(int x, int y) {
-		//TODO
+	public void scroll(final int x, final int y) {
+		if(EventQueue.isDispatchThread()) {
+			this.scrollImpl(x, y);
+		}
+		else {
+			EventQueue.invokeLater(new Runnable() {
+				public void run() {
+					scrollImpl(x, y);					
+				}
+			});
+		}
 	}
 
+	private void scrollImpl(int x, int y) {
+		this.scrollTo(new Rectangle(x, y, 16, 16), false, false);		
+	}
+	
 	/**
 	 * Clears the current document if any.
 	 * If called outside the GUI thread, the operation
@@ -205,23 +256,7 @@ public class HtmlPanel extends JComponent implements FrameContext {
 		this.revalidate();
 		this.repaint();
 	}
-	
-	/**
-	 * Sets an HTML DOM node and invalidates the component so it is
-	 * rendered immediately in the GUI thread.
-	 * @param node This should
-	 * normally be a Document instance obtained with
-	 * {@link org.lobobrowser.html.parser.DocumentBuilderImpl}.
-	 * <p>
-	 * Note: It is safe to call this method outside of the GUI thread.
-	 * @param rcontext A renderer context.
-	 * @param pcontext A parser context.
-	 * @deprecated HtmlParserContext is no longer used here.
-	 */
-	public void setDocument(final Document node, final HtmlRendererContext rcontext, final HtmlParserContext pcontext) {
-		this.setDocument(node, rcontext);
-	}
-	
+		
 	/**
 	 * Sets an HTML DOM node and invalidates the component so it is
 	 * rendered as soon as possible in the GUI thread. 
@@ -250,6 +285,39 @@ public class HtmlPanel extends JComponent implements FrameContext {
 		}
 	}
 
+	/**
+	 * Scrolls to the element identified by the given ID in
+	 * the current document. 
+	 * <p>
+	 * If this method is invoked outside the GUI thread, 
+	 * the operation is scheduled to be performed as soon
+	 * as possible in the GUI thread.
+	 * @param nameOrId The name or ID of the element in the document.
+	 */
+	public void scrollToElement(final String nameOrId) {
+		if(EventQueue.isDispatchThread()) {
+			this.scrollToElementImpl(nameOrId);
+		}
+		else {
+			EventQueue.invokeLater(new Runnable() {
+				public void run() {
+					scrollToElementImpl(nameOrId);
+				}
+			});
+		}
+	}
+
+	private void scrollToElementImpl(String nameOrId) {
+		NodeImpl node = this.rootNode;
+		if(node instanceof HTMLDocumentImpl) {
+			HTMLDocumentImpl doc = (HTMLDocumentImpl) node;
+			org.w3c.dom.Element element = doc.getElementById(nameOrId);
+			if(element != null) {
+				this.scrollTo(element);
+			}
+		}
+	}
+	
 	private void setDocumentImpl(Document node, HtmlRendererContext rcontext) {
 		// Expected to be called in the GUI thread.
 		if(!(node instanceof HTMLDocumentImpl)) {
@@ -293,6 +361,9 @@ public class HtmlPanel extends JComponent implements FrameContext {
 
 	/**
 	 * Renders HTML given as a string.
+	 * @param htmlSource The HTML source code.
+	 * @param uri A base URI used to resolve item URIs.
+	 * @param rcontext The {@link HtmlRendererContext} instance.
 	 */
 	public void setHtml(String htmlSource, String uri, HtmlRendererContext rcontext) {
 		try {
@@ -468,6 +539,9 @@ public class HtmlPanel extends JComponent implements FrameContext {
 	 * the selection.
 	 * <p>
 	 * Note: This method should be invoked in the GUI thread.
+	 * @return A node enclosing the current selection, or <code>null</code>
+	 *         if there is no such node. It also returns <code>null</code>
+	 *         for FRAMESETs.
 	 */
 	public org.w3c.dom.Node getSelectionNode() {
 		HtmlBlockPanel block = this.htmlBlock;
@@ -624,6 +698,10 @@ public class HtmlPanel extends JComponent implements FrameContext {
 		public void nodeLoaded(NodeImpl node) {
 			HtmlPanel.this.addNotification(new DocumentNotification(DocumentNotification.GENERIC, node));
 		}
+
+		public void structureInvalidated(NodeImpl node) {
+			HtmlPanel.this.addNotification(new DocumentNotification(DocumentNotification.GENERIC, node));
+		}
 	}
 	
 	private class NotificationTimerAction implements java.awt.event.ActionListener {
