@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Aug 28, 2005
@@ -51,8 +51,8 @@ public class HtmlParser {
 	private final String publicId;
 	private final String systemId;
 	
-	private static final Map ENTITIES = new HashMap();
-	private static final Map ELEMENT_INFOS = new HashMap();
+	private static final Map ENTITIES = new HashMap(256);
+	private static final Map ELEMENT_INFOS = new HashMap(35);
 
 	/**
 	 * A node <code>UserData</code> key used to tell
@@ -177,9 +177,173 @@ public class HtmlParser {
 		entities.put("thorn", new Character((char) 254));
 		entities.put("yuml", new Character((char) 255));
 
-		//TODO: See http://elcursillo0.tripod.com/muestras/caracters.html
-		//TODO: See http://www.koders.com/java/fid0B623B057D114F8A68B54C6060C2B1AEE2D04750.aspx
+		//symbols from http://de.selfhtml.org/html/referenz/zeichen.htm
 		
+		//greek letters
+		entities.put("Alpha", new Character((char) 913));
+		entities.put("Beta", new Character((char) 914));
+		entities.put("Gamma", new Character((char) 915));
+		entities.put("Delta", new Character((char) 916));
+		entities.put("Epsilon", new Character((char) 917));
+		entities.put("Zeta", new Character((char) 918));
+		entities.put("Eta", new Character((char) 919));
+		entities.put("Theta", new Character((char) 920));
+		entities.put("Iota", new Character((char) 921));
+		entities.put("Kappa", new Character((char) 922));
+		entities.put("Lambda", new Character((char) 923));
+		entities.put("Mu", new Character((char) 924));
+		entities.put("Nu", new Character((char) 925));
+		entities.put("Xi", new Character((char) 926));
+		entities.put("Omicron", new Character((char) 927));
+		entities.put("Pi", new Character((char) 928));
+		entities.put("Rho", new Character((char) 929));
+		entities.put("Sigma", new Character((char) 930));
+		entities.put("Sigmaf", new Character((char) 931));
+		entities.put("Tau", new Character((char) 932));
+		entities.put("Upsilon", new Character((char) 933));
+		entities.put("Phi", new Character((char) 934));
+		entities.put("Chi", new Character((char) 935));
+		entities.put("Psi", new Character((char) 936));
+		entities.put("Omega", new Character((char) 937));
+		
+		entities.put("alpha", new Character((char) 945));
+		entities.put("beta", new Character((char) 946));
+		entities.put("gamma", new Character((char) 947));
+		entities.put("delta", new Character((char) 948));
+		entities.put("epsilon", new Character((char) 949));
+		entities.put("zeta", new Character((char) 950));
+		entities.put("eta", new Character((char) 951));
+		entities.put("theta", new Character((char) 952));
+		entities.put("iota", new Character((char) 953));
+		entities.put("kappa", new Character((char) 954));
+		entities.put("lambda", new Character((char) 955));
+		entities.put("mu", new Character((char) 956));
+		entities.put("nu", new Character((char) 957));
+		entities.put("xi", new Character((char) 958));
+		entities.put("omicron", new Character((char) 959));
+		entities.put("pi", new Character((char) 960));
+		entities.put("rho", new Character((char) 961));
+		entities.put("sigma", new Character((char) 962));
+		entities.put("sigmaf", new Character((char) 963));
+		entities.put("tau", new Character((char) 964));
+		entities.put("upsilon", new Character((char) 965));
+		entities.put("phi", new Character((char) 966));
+		entities.put("chi", new Character((char) 967));
+		entities.put("psi", new Character((char) 968));
+		entities.put("omega", new Character((char) 969));
+		entities.put("thetasym", new Character((char) 977));
+		entities.put("upsih", new Character((char) 978));
+		entities.put("piv", new Character((char) 982));
+		
+		//math symbols
+		entities.put("forall", new Character((char) 8704));
+		entities.put("part", new Character((char) 8706));
+		entities.put("exist", new Character((char) 8707));
+		entities.put("empty", new Character((char) 8709));
+		entities.put("nabla", new Character((char) 8711));
+		entities.put("isin", new Character((char) 8712));
+		entities.put("notin", new Character((char) 8713));
+		entities.put("ni", new Character((char) 8715));
+		entities.put("prod", new Character((char) 8719));
+		entities.put("sum", new Character((char) 8721));
+		entities.put("minus", new Character((char) 8722));
+		entities.put("lowast", new Character((char) 8727));
+		entities.put("radic", new Character((char) 8730));
+		entities.put("prop", new Character((char) 8733));
+		entities.put("infin", new Character((char) 8734));
+		entities.put("ang", new Character((char) 8736));
+		entities.put("and", new Character((char) 8743));
+		entities.put("or", new Character((char) 8744));
+		entities.put("cap", new Character((char) 8745));
+		entities.put("cup", new Character((char) 8746));
+		entities.put("int", new Character((char) 8747));
+		entities.put("there4", new Character((char) 8756));
+		entities.put("sim", new Character((char) 8764));
+		entities.put("cong", new Character((char) 8773));
+		entities.put("asymp", new Character((char) 8776));
+		entities.put("ne", new Character((char) 8800));
+		entities.put("equiv", new Character((char) 8801));
+		entities.put("le", new Character((char) 8804));
+		entities.put("ge", new Character((char) 8805));
+		entities.put("sub", new Character((char) 8834));
+		entities.put("sup", new Character((char) 8835));
+		entities.put("nsub", new Character((char) 8836));
+		entities.put("sube", new Character((char) 8838));
+		entities.put("supe", new Character((char) 8839));
+		entities.put("oplus", new Character((char) 8853));
+		entities.put("otimes", new Character((char) 8855));
+		entities.put("perp", new Character((char) 8869));
+		entities.put("sdot", new Character((char) 8901));
+		entities.put("loz", new Character((char) 9674));
+		
+		//technical symbols
+		entities.put("lceil", new Character((char) 8968));
+		entities.put("rceil", new Character((char) 8969));
+		entities.put("lfloor", new Character((char) 8970));
+		entities.put("rfloor", new Character((char) 8971));
+		entities.put("lang", new Character((char) 9001));
+		entities.put("rang", new Character((char) 9002));
+		
+		//arrow symbols
+		entities.put("larr", new Character((char) 8592));
+		entities.put("uarr", new Character((char) 8593));
+		entities.put("rarr", new Character((char) 8594));
+		entities.put("darr", new Character((char) 8595));
+		entities.put("harr", new Character((char) 8596));
+		entities.put("crarr", new Character((char) 8629));
+		entities.put("lArr", new Character((char) 8656));
+		entities.put("uArr", new Character((char) 8657));
+		entities.put("rArr", new Character((char) 8658));
+		entities.put("dArr", new Character((char) 8659));
+		entities.put("hArr", new Character((char) 8960));
+		
+		//divers symbols
+		entities.put("bull", new Character((char) 8226));
+		entities.put("prime", new Character((char) 8242));
+		entities.put("Prime", new Character((char) 8243));
+		entities.put("oline", new Character((char) 8254));
+		entities.put("weierp", new Character((char) 8472));
+		entities.put("image", new Character((char) 8465));
+		entities.put("real", new Character((char) 8476));
+		entities.put("trade", new Character((char) 8482));
+		entities.put("euro", new Character((char) 8364));
+		entities.put("alefsym", new Character((char) 8501));
+		entities.put("spades", new Character((char) 9824));
+		entities.put("clubs", new Character((char) 9827));
+		entities.put("hearts", new Character((char) 9829));
+		entities.put("diams", new Character((char) 9830));
+
+		//ext lat symbols
+		entities.put("OElig", new Character((char) 338));
+		entities.put("oelig", new Character((char) 339));
+		entities.put("Scaron", new Character((char) 352));
+		entities.put("scaron", new Character((char) 353));
+		entities.put("fnof", new Character((char) 402));
+
+		//interpunction
+		entities.put("ensp", new Character((char) 8194));
+		entities.put("emsp", new Character((char) 8195));
+		entities.put("thinsp", new Character((char) 8201));
+		entities.put("zwnj", new Character((char) 8204));
+		entities.put("zwj", new Character((char) 8205));
+		entities.put("lrm", new Character((char) 8206));
+		entities.put("rlm", new Character((char) 8207));
+		
+		entities.put("sbquo", new Character((char) 8218));
+		entities.put("ldquo", new Character((char) 8220));
+		entities.put("rdquo", new Character((char) 8221));
+		entities.put("bdquo", new Character((char) 8222));
+		entities.put("dagger", new Character((char) 8224));
+		entities.put("Dagger", new Character((char) 8225));
+		entities.put("hellip", new Character((char) 8230));
+		entities.put("permil", new Character((char) 8240));
+		entities.put("lsaquo", new Character((char) 8249));
+		entities.put("rsaquo", new Character((char) 8250));
+		
+		//diacrit symb
+		entities.put("circ", new Character((char) 710));
+		entities.put("tilde", new Character((char) 732));
+				
 		Map elementInfos = ELEMENT_INFOS;
 		
 		elementInfos.put("NOSCRIPT", new ElementInfo(true, ElementInfo.END_ELEMENT_REQUIRED, null, true));
@@ -246,25 +410,6 @@ public class HtmlParser {
 		//TODO: Keep adding tags here		
 	}
 
-	/**
-	 * Constructs a <code>HtmlParser</code>.
-	 * @param context An instance of {@link org.lobobrowser.html.HtmlRendererContext},
-	 * which may be an instance of {@link org.lobobrowser.html.test.SimpleHtmlRendererContext}.
-	 * @param document An instanceof of <code>HTMLDocument</code>.
-	 * @param errorHandler The error handler. 
-	 * @param publicId The public ID of the document.
-	 * @param systemId The system ID of the document.
-	 * @deprecated HtmlParserContext not needed by parser
-	 *             and UserAgentContext should be passed instead.
-	 */
-	public HtmlParser(HtmlParserContext context, HTMLDocument document, ErrorHandler errorHandler, String publicId, String systemId) {
-		this.ucontext = null;
-		this.document = document;
-		this.errorHandler = errorHandler;
-		this.publicId = publicId;
-		this.systemId = systemId;
-	}
-	
 	/**
 	 * Constructs a <code>HtmlParser</code>.
 	 * @param document An instanceof of <code>HTMLDocument</code>.
@@ -453,6 +598,12 @@ public class HtmlParser {
 					this.passEndOfTag(reader);
 					return TOKEN_END_ELEMENT;
 				}
+				else if (tag.startsWith("?")) {					
+					tag = tag.substring(1);
+					StringBuffer data = readProcessingInstruction(reader);					
+					parent.appendChild(doc.createProcessingInstruction(tag, data.toString()));					
+					return TOKEN_FULL_ELEMENT;
+				}
 				else {
 					Element element = doc.createElement(tag);
 					element.setUserData(MODIFYING_KEY, Boolean.TRUE, null);
@@ -883,6 +1034,21 @@ public class HtmlParser {
 		}
 	}
 	
+	private final StringBuffer readProcessingInstruction(LineNumberReader reader) throws IOException {
+		StringBuffer pidata = new StringBuffer();		
+		if(this.justReadTagEnd) {
+			return pidata;
+		}
+		int ch;
+		for(ch = reader.read();ch != -1 && ch != '>'; ch = reader.read())
+		{			
+			pidata.append((char)ch);			
+		}		
+		this.justReadTagBegin = false;
+		this.justReadTagEnd = ch != -1;
+		return pidata;
+	}
+	
 	private final boolean readAttribute(LineNumberReader reader, Element element) throws IOException, SAXException {
 		if(this.justReadTagEnd) {
 			return false;
