@@ -1,4 +1,4 @@
-/*    GNU LESSER GENERAL PUBLIC LICENSE
+/*  GNU LESSER GENERAL PUBLIC LICENSE
     Copyright (C) 2006 The Lobo Project
 
     This library is free software; you can redistribute it and/or
@@ -15,15 +15,13 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 
 package org.lobobrowser.html.renderer;
 
 import java.awt.*;
 import java.awt.image.ImageObserver;
-import java.net.MalformedURLException;
-import java.net.URL;
 import java.security.AccessController;
 import java.security.PrivilegedAction;
 import java.util.Collection;
@@ -37,7 +35,7 @@ import org.lobobrowser.html.ReadyStateChangeListener;
 import org.lobobrowser.html.UserAgentContext;
 import org.lobobrowser.html.domimpl.*;
 import org.lobobrowser.html.style.BackgroundInfo;
-import org.lobobrowser.html.style.CSS2PropertiesImpl;
+import org.lobobrowser.html.style.AbstractCSS2Properties;
 import org.lobobrowser.html.style.HtmlValues;
 import org.lobobrowser.html.style.RenderState;
 import org.lobobrowser.util.gui.ColorFactory;
@@ -51,16 +49,7 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 	public static final int OVERFLOW_VISIBLE = 4;
 	public static final int OVERFLOW_VERTICAL = 5;
 	
-	public static final int BORDER_STYLE_NONE = 0;
-	public static final int BORDER_STYLE_HIDDEN = 1;
-	public static final int BORDER_STYLE_DOTTED = 2;
-	public static final int BORDER_STYLE_DASHED = 3;
-	public static final int BORDER_STYLE_SOLID = 4;
-	public static final int BORDER_STYLE_DOUBLE = 5;
-	public static final int BORDER_STYLE_GROOVE = 6;
-	public static final int BORDER_STYLE_RIDGE = 7;
-	public static final int BORDER_STYLE_INSET = 8;
-	public static final int BORDER_STYLE_OUTSET = 9;
+	protected static final Integer INVALID_SIZE = new Integer(Integer.MIN_VALUE);
 	
 	/**
 	 * A collection of all GUI components
@@ -134,9 +123,29 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 
 	protected void invalidateLayoutLocal() {
 		this.cachedOverflow = -1;
+		this.declaredWidth = INVALID_SIZE;
+		this.declaredHeight = INVALID_SIZE;
+		this.lastAvailHeightForDeclared = -1;
+		this.lastAvailWidthForDeclared = -1;
 	}
 
-	protected int getDeclaredWidth(RenderState renderState, int availWidth) {
+	private Integer declaredWidth = INVALID_SIZE;
+	private Integer declaredHeight = INVALID_SIZE;
+	private int lastAvailWidthForDeclared = -1;
+	private int lastAvailHeightForDeclared = -1;
+	
+	protected Integer getDeclaredWidth(RenderState renderState, int availWidth) {
+		Integer dw = this.declaredWidth;
+		if(dw == INVALID_SIZE || availWidth != this.lastAvailWidthForDeclared) {
+			this.lastAvailWidthForDeclared = availWidth;
+			int dwInt = this.getDeclaredWidthImpl(renderState, availWidth);
+			dw = dwInt == -1 ? null : new Integer(dwInt);
+			this.declaredWidth  = dw;
+		}
+		return dw;
+	}
+		
+	private final int getDeclaredWidthImpl(RenderState renderState, int availWidth) {
 		Object rootNode = this.modelNode;
 		if(rootNode instanceof HTMLElementImpl) {
 			HTMLElementImpl element = (HTMLElementImpl) rootNode;
@@ -155,7 +164,18 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 		}
 	}
 
-	protected int getDeclaredHeight(RenderState renderState, int availHeight) {
+	protected Integer getDeclaredHeight(RenderState renderState, int availHeight) {
+		Integer dh = this.declaredHeight;
+		if(dh == INVALID_SIZE || availHeight != this.lastAvailHeightForDeclared) {
+			this.lastAvailHeightForDeclared = availHeight;
+			int dhInt = this.getDeclaredHeightImpl(renderState, availHeight);
+			dh = dhInt == -1 ? null : new Integer(dhInt);
+			this.declaredHeight  = dh;
+		}
+		return dh;		
+	}
+	
+	protected int getDeclaredHeightImpl(RenderState renderState, int availHeight) {
 		Object rootNode = this.modelNode;
 		if(rootNode instanceof HTMLElementImpl) {
 			HTMLElementImpl element = (HTMLElementImpl) rootNode;
@@ -227,26 +247,21 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 	public void paint(Graphics g) {
 	}
 
-	public final void layout(int availWidth, int availHeight) {
-		this.layout(availWidth, availHeight, false, false);
-	}
-
 	/**
 	 * Lays out children, and deals with "valid" state. Override doLayout method
 	 * instead of this one.
 	 */
-	public final void layout(int availWidth, int availHeight, boolean expandWidth, boolean expandHeight) {
+	public final void layout(int availWidth, int availHeight) {
 		// Must call doLayout regardless of validity state.
 		try {
-			this.doLayout(availWidth, availHeight, expandWidth, expandHeight);	
+			this.doLayout(availWidth, availHeight);	
 		} finally {
 			this.layoutUpTreeCanBeInvalidated = true;
 			this.layoutDeepCanBeInvalidated = true;
-//			this.renderStyleCanBeInvalidated = true;
 		}
 	}	
 	
-	protected abstract void doLayout(int availWidth, int availHeight, boolean expandWidth, boolean expandHeight);	
+	protected abstract void doLayout(int availWidth, int availHeight);	
 	
 	protected final void sendGUIComponentsToParent() {
 		// Ensures that parent has all the components
@@ -258,7 +273,7 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 			Iterator i = gc.iterator();
 			while(i.hasNext()) {
 				count++;
-				rc.add((Component) i.next());
+				rc.addComponent((Component) i.next());
 			}
 		}		
 	}
@@ -273,7 +288,7 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 	/* (non-Javadoc)
 	 * @see org.xamjwg.html.renderer.RenderableContainer#add(java.awt.Component)
 	 */
-	public Component add(Component component) {
+	public Component addComponent(Component component) {
 		// Expected to be called in GUI thread.
 		// Adds only in local collection.
 		// Does not remove from parent.
@@ -337,46 +352,42 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 			this.lastBackgroundImageUri = backgroundImageUri;
 			this.loadBackgroundImage(backgroundImageUri);
 		}
-		CSS2PropertiesImpl props = rootElement.getCurrentStyle();
+		AbstractCSS2Properties props = rootElement.getCurrentStyle();
 		if(props == null) {
 			this.clearStyle();
 		}
 		else {
-			this.borderInsets = null;
-			this.borderStyles = null;
-			this.borderTopColor = null;
-			this.borderLeftColor = null;
-			this.borderBottomColor = null;
-			this.borderRightColor = null;
-			String border = props.getBorder();
-			if(border != null) {
-				this.applyBorder(rs, border);
-			}
-			this.borderInsets = HtmlValues.getBorderInsets(this.borderInsets, props, rs);
-			String borderColorText = props.getBorderColor();
-			if(borderColorText != null) {
-				Color[] colorsArray = HtmlValues.getColors(borderColorText);
-				this.borderTopColor = colorsArray[0];
-				this.borderLeftColor = colorsArray[1];
-				this.borderBottomColor = colorsArray[2];
-				this.borderRightColor = colorsArray[3];
-			}
+			Insets borderStyles = HtmlValues.getBorderStyles(props);
+			this.borderStyles = borderStyles;
+			this.borderInsets = HtmlValues.getBorderInsets(borderStyles, props, rs);
 			String borderTopColorText = props.getBorderTopColor();
 			if(borderTopColorText != null) {
 				this.borderTopColor = ColorFactory.getInstance().getColor(borderTopColorText);
 			}
+			else {
+				this.borderTopColor = null;
+			}
 			String borderLeftColorText = props.getBorderLeftColor();
 			if(borderLeftColorText != null) {
 				this.borderLeftColor = ColorFactory.getInstance().getColor(borderLeftColorText);
 			}
+			else {
+				this.borderLeftColor = null;
+			}
 			String borderBottomColorText = props.getBorderBottomColor();
 			if(borderBottomColorText != null) {
 				this.borderBottomColor = ColorFactory.getInstance().getColor(borderBottomColorText);
 			}
+			else {
+				this.borderBottomColor = null;
+			}
 			String borderRightColorText = props.getBorderRightColor();
 			if(borderRightColorText != null) {
 				this.borderRightColor = ColorFactory.getInstance().getColor(borderRightColorText);
 			}
+			else {
+				this.borderRightColor = null;
+			}
 			String zIndex = props.getZIndex();
 			if(zIndex != null) {
 				try {
@@ -420,14 +431,24 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 			});
 			SecurityManager sm = System.getSecurityManager();
 			if(sm == null) {
-				request.open("GET", imageURL);					
+				try {	
+					request.open("GET", imageURL);	
+					request.send(null);
+				} catch(java.io.IOException thrown) {
+					logger.log(Level.WARNING, "loadBackgroundImage()", thrown);
+				}
 			}
 			else {
 				AccessController.doPrivileged(new PrivilegedAction() {
 					public Object run() {
 						// Code might have restrictions on accessing
 						// items from elsewhere.
-						request.open("GET", imageURL);
+						try {
+							request.open("GET", imageURL);
+							request.send(null);
+						} catch(java.io.IOException thrown) {
+							logger.log(Level.WARNING, "loadBackgroundImage()", thrown);
+						}
 						return null;
 					}
 				});
@@ -439,46 +460,46 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 		return this.zIndex;
 	}
 
-	void applyBorder(RenderState renderState, String border) {
-		String[] tokens = HtmlValues.splitCssValue(border);
-		for(int i = 0; i < tokens.length; i++) {
-			String token = tokens[i];
-			if(HtmlValues.isLength(token)) {
-				int pixelSize = HtmlValues.getPixelSize(token, renderState, 0);
-				Insets bi = new Insets(pixelSize, pixelSize, pixelSize, pixelSize);
-				this.borderInsets = bi;
-			}
-			else if(ColorFactory.getInstance().isColor(token)) {
-				Color color = ColorFactory.getInstance().getColor(token);
-				this.borderLeftColor = color;
-				this.borderRightColor = color;
-				this.borderTopColor = color;
-				this.borderBottomColor = color;
-			}
-			else if(HtmlValues.isBorderStyle(token)) {
-				//TODO: dotted and so forth
-				if("solid".equalsIgnoreCase(token)) {
-					Insets bi = this.borderInsets;
-					if(bi == null) {
-						bi = new Insets(4, 4, 4, 4);
-						this.borderInsets = bi;
-					}
-					this.borderStyles = new Insets(BORDER_STYLE_SOLID,BORDER_STYLE_SOLID,BORDER_STYLE_SOLID,BORDER_STYLE_SOLID);
-				}
-				else if("dashed".equalsIgnoreCase(token)) {
-					Insets bi = this.borderInsets;
-					if(bi == null) {
-						bi = new Insets(4, 4, 4, 4);
-						this.borderInsets = bi;
-					}
-					this.borderStyles = new Insets(BORDER_STYLE_DASHED,BORDER_STYLE_DASHED,BORDER_STYLE_DASHED,BORDER_STYLE_DASHED);
-				}
-				else {
-					this.borderStyles = null;					
-				}
-			}
-		}
-	}
+//	void applyBorder(RenderState renderState, String border) {
+//		String[] tokens = HtmlValues.splitCssValue(border);
+//		for(int i = 0; i < tokens.length; i++) {
+//			String token = tokens[i];
+//			if(HtmlValues.isLength(token)) {
+//				int pixelSize = HtmlValues.getPixelSize(token, renderState, 0);
+//				Insets bi = new Insets(pixelSize, pixelSize, pixelSize, pixelSize);
+//				this.borderInsets = bi;
+//			}
+//			else if(ColorFactory.getInstance().isColor(token)) {
+//				Color color = ColorFactory.getInstance().getColor(token);
+//				this.borderLeftColor = color;
+//				this.borderRightColor = color;
+//				this.borderTopColor = color;
+//				this.borderBottomColor = color;
+//			}
+//			else if(HtmlValues.isBorderStyle(token)) {
+//				//TODO: dotted and so forth
+//				if("solid".equalsIgnoreCase(token)) {
+//					Insets bi = this.borderInsets;
+//					if(bi == null) {
+//						bi = new Insets(4, 4, 4, 4);
+//						this.borderInsets = bi;
+//					}
+//					this.borderStyles = new Insets(HtmlValues.BORDER_STYLE_SOLID,HtmlValues.BORDER_STYLE_SOLID,HtmlValues.BORDER_STYLE_SOLID,HtmlValues.BORDER_STYLE_SOLID);
+//				}
+//				else if("dashed".equalsIgnoreCase(token)) {
+//					Insets bi = this.borderInsets;
+//					if(bi == null) {
+//						bi = new Insets(4, 4, 4, 4);
+//						this.borderInsets = bi;
+//					}
+//					this.borderStyles = new Insets(HtmlValues.BORDER_STYLE_DASHED,HtmlValues.BORDER_STYLE_DASHED,HtmlValues.BORDER_STYLE_DASHED,HtmlValues.BORDER_STYLE_DASHED);
+//				}
+//				else {
+//					this.borderStyles = null;					
+//				}
+//			}
+//		}
+//	}
 
 	private Color getBorderTopColor() {
 		Color c = this.borderTopColor;
@@ -536,11 +557,11 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 				Insets borderStyles = this.borderStyles;
 				if(btop > 0) {
 					g.setColor(this.getBorderTopColor());
-					int borderStyle = borderStyles == null ? BORDER_STYLE_SOLID : borderStyles.top;
+					int borderStyle = borderStyles == null ? HtmlValues.BORDER_STYLE_SOLID : borderStyles.top;
 					for(int i = 0; i < btop; i++) {
 						int leftOffset = (i * bleft) / btop;
 						int rightOffset = (i * bright) / btop;
-						if(borderStyle == BORDER_STYLE_DASHED) {
+						if(borderStyle == HtmlValues.BORDER_STYLE_DASHED) {
 							GUITasks.drawDashed(g, startX + leftOffset, startY + i, startX + totalWidth - rightOffset - 1, startY + i, 10 + btop, 6);
 						}
 						else {
@@ -549,13 +570,13 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 					}
 				}
 				if(bright > 0) {
-					int borderStyle = borderStyles == null ? BORDER_STYLE_SOLID : borderStyles.right;
+					int borderStyle = borderStyles == null ? HtmlValues.BORDER_STYLE_SOLID : borderStyles.right;
 					g.setColor(this.getBorderRightColor());
 					int lastX = startX + totalWidth - 1;
 					for(int i = 0; i < bright; i++) {
 						int topOffset = (i * btop) / bright;
 						int bottomOffset = (i * bbottom) / bright;
-						if(borderStyle == BORDER_STYLE_DASHED) {
+						if(borderStyle == HtmlValues.BORDER_STYLE_DASHED) {
 							GUITasks.drawDashed(g, lastX - i, startY + topOffset, lastX - i, startY + totalHeight - bottomOffset - 1, 10 + bright, 6);
 						}
 						else {
@@ -564,13 +585,13 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 					}				
 				}
 				if(bbottom > 0) {
-					int borderStyle = borderStyles == null ? BORDER_STYLE_SOLID : borderStyles.bottom;
+					int borderStyle = borderStyles == null ? HtmlValues.BORDER_STYLE_SOLID : borderStyles.bottom;
 					g.setColor(this.getBorderBottomColor());
 					int lastY = startY + totalHeight - 1;
 					for(int i = 0; i < bbottom; i++) {
 						int leftOffset = (i * bleft) / bbottom;
 						int rightOffset = (i * bright) / bbottom;					
-						if(borderStyle == BORDER_STYLE_DASHED) {
+						if(borderStyle == HtmlValues.BORDER_STYLE_DASHED) {
 							GUITasks.drawDashed(g, startX + leftOffset, lastY - i, startX + totalWidth - rightOffset - 1, lastY - i, 10 + bbottom, 6);
 						}
 						else {
@@ -579,12 +600,12 @@ abstract class BaseElementRenderable extends BaseRCollection implements RElement
 					}				
 				}
 				if(bleft > 0) {
-					int borderStyle = borderStyles == null ? BORDER_STYLE_SOLID : borderStyles.left;
+					int borderStyle = borderStyles == null ? HtmlValues.BORDER_STYLE_SOLID : borderStyles.left;
 					g.setColor(this.getBorderLeftColor());
 					for(int i = 0; i < bleft; i++) {
 						int topOffset = (i * btop) / bleft;
 						int bottomOffset = (i * bbottom) / bleft;
-						if(borderStyle == BORDER_STYLE_DASHED) {
+						if(borderStyle == HtmlValues.BORDER_STYLE_DASHED) {
 							GUITasks.drawDashed(g, startX + i, startY + topOffset, startX + i, startY + totalHeight - bottomOffset - 1, 10 + bleft, 6);
 						}
 						else {
