@@ -16,22 +16,23 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 
 package org.lobobrowser.html.renderer;
 
 import java.util.Iterator;
 
-interface RCollection extends BoundableRenderable {
+/**
+ * A {@link Renderable} with children.
+ */
+public interface RCollection extends BoundableRenderable {
+	/**
+	 * Gets the collection of {@link Renderable} chjldren.
+	 */
 	public Iterator getRenderables();
 	public void updateWidgetBounds(int guiX, int guiY);
-	
-//	/**
-//	 * Invalidates RenderState in all descendents.
-//	 */
-//	public void invalidateRenderStyle();
-//	
+
 	/**
 	 * Invalidates layout in all descendents.
 	 */
