@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Apr 16, 2005
@@ -25,7 +25,6 @@ package org.lobobrowser.html.renderer;
 
 import java.awt.*;
 import java.util.*;
-
 import org.lobobrowser.html.domimpl.ModelNode;
 import org.lobobrowser.html.style.RenderState;
 
@@ -56,6 +55,15 @@ class RLine extends BaseRCollection {
 		this.layoutUpTreeCanBeInvalidated = true;
 	}
 	
+	/**
+	 * This method should only be invoked when the line has
+	 * no items yet.
+	 */
+	public void changeLimits(int x, int desiredMaxWidth) {
+		this.x = x;
+		this.desiredMaxWidth = desiredMaxWidth;
+	}
+	
 	public int getBaselineOffset() {
 		return this.baseLineOffset;
 	}
@@ -109,21 +117,25 @@ class RLine extends BaseRCollection {
 	public final void addStyleChanger(RStyleChanger sc) {
 		this.renderables.add(sc);
 	}
-	
+
+	public final void simplyAdd(Renderable r) {
+		this.renderables.add(r);
+	}
+
 	/**
 	 * This method adds and positions a renderable in the line, if possible.
 	 * Note that RLine does not set sizes, but only origins.
 	 * @throws OverflowException Thrown if the renderable overflows the line. All overflowing renderables are added to the exception.
 	 */
-	public final void add(Renderable renderable, boolean allowOverflow) throws OverflowException {
+	public final void add(Renderable renderable, int overflowType) throws OverflowException {
 		if(renderable instanceof RWord) {
-			this.addWord((RWord) renderable, allowOverflow);
+			this.addWord((RWord) renderable, overflowType);
 		}
 		else if(renderable instanceof RBlank) {
 			this.addBlank((RBlank) renderable);
 		}
-		else if(renderable instanceof RElement) {
-			this.addElement((RElement) renderable, allowOverflow);
+		else if(renderable instanceof RUIControl) {
+			this.addElement((RElement) renderable, overflowType);
 		}
 		else if(renderable instanceof RSpacing) {
 			this.addSpacing((RSpacing) renderable);
@@ -131,16 +143,23 @@ class RLine extends BaseRCollection {
 		else if(renderable instanceof RStyleChanger) {
 			this.addStyleChanger((RStyleChanger) renderable);
 		}
+		else if(renderable instanceof RFloatInfo) {
+			this.simplyAdd((RFloatInfo) renderable);
+		}
 		else {
 			throw new IllegalArgumentException("Can't add " + renderable);
 		}
 	}
 	
-	public final void addWord(RWord rword, boolean allowOverflow) throws OverflowException {
+	public static final int OVERFLOW_ALLOW_ALL = 0;
+	public static final int OVERFLOW_ALLOW_FIRST = 1;
+	public static final int OVERFLOW_ALLOW_NONE = 2;
+		
+	public final void addWord(RWord rword, int overflowType) throws OverflowException {
 		// Check if it fits horzizontally
 		int offset = this.xoffset;
 		int wiwidth = rword.width;
-		if(!allowOverflow && (offset != 0 && offset + wiwidth > this.desiredMaxWidth)) {
+		if(overflowType != OVERFLOW_ALLOW_ALL && (offset != 0 || overflowType == OVERFLOW_ALLOW_NONE) && (offset + wiwidth > this.desiredMaxWidth)) {
 			ArrayList renderables = this.renderables;
 			ArrayList overflow = null;
 			boolean cancel = false;
@@ -272,15 +291,12 @@ class RLine extends BaseRCollection {
 		relement.setY(yoffset);
 	}
 	
-	public final void addElement(RElement relement, boolean allowOverflow) throws OverflowException {
+	private final void addElement(RElement relement, int overflowType) throws OverflowException {
 		// Check if it fits horizontally
 		int origXOffset = this.xoffset;
 		int desiredMaxWidth = this.desiredMaxWidth;
-//		int componentHeight = this.availHeight;
-//      (already layed out)
-//		relement.layout(desiredMaxWidth, componentHeight);
 		int pw = relement.getWidth();
-		if(!allowOverflow && (origXOffset != 0 && origXOffset + pw > desiredMaxWidth)) {
+		if(overflowType != OVERFLOW_ALLOW_ALL && (origXOffset != 0 || overflowType == OVERFLOW_ALLOW_NONE) && (origXOffset + pw > desiredMaxWidth)) {
 			throw new OverflowException(Collections.singleton(relement));
 		}
 		//Note: Renderable for widget doesn't paint the widget, but
@@ -594,4 +610,18 @@ class RLine extends BaseRCollection {
 	public boolean isContainedByNode() {
 		return false;
 	}
+	
+	private LineBreak lineBreak;
+
+	public LineBreak getLineBreak() {
+		return lineBreak;
+	}
+
+	public void setLineBreak(LineBreak lineBreak) {
+		this.lineBreak = lineBreak;
+	}
+	
+	public boolean isEmpty() {
+		return this.xoffset == 0;
+	}
 }
