@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 
 package org.lobobrowser.html.style;
@@ -39,8 +39,6 @@ import org.w3c.dom.stylesheets.*;
  * instances. The instances inserted in that list
  * are obtained by means of the {@link #getStyleDeclarations(HTMLElementImpl, String, String, String)}
  * method.
- * @author user
- *
  */
 public class StyleSheetAggregator {
 	private final HTMLDocumentImpl document;
@@ -73,46 +71,44 @@ public class StyleSheetAggregator {
 		HTMLDocumentImpl document = this.document;
 		if(rule instanceof CSSStyleRule) {
 			CSSStyleRule sr = (CSSStyleRule) rule;
-			String selectorText = sr.getSelectorText();
-			StringTokenizer commaTok = new StringTokenizer(selectorText, ",");
+			String selectorList = sr.getSelectorText();
+			StringTokenizer commaTok = new StringTokenizer(selectorList, ",");
 			while(commaTok.hasMoreTokens()) {
 				String selectorPart = commaTok.nextToken().toLowerCase();
-				ArrayList ancestorSelectors = null;
-				String selector = null;
+				ArrayList simpleSelectors = null;
+				String lastSelectorText = null;
 				StringTokenizer tok = new StringTokenizer(selectorPart, " \t\r\n");
-				while(tok.hasMoreTokens()) {
-					String token = tok.nextToken();
-					if(tok.hasMoreTokens()) {
-						if(ancestorSelectors == null) {
-							ancestorSelectors = new ArrayList();
+				if(tok.hasMoreTokens()) {
+					simpleSelectors = new ArrayList();
+					for(;;) {
+						String token = tok.nextToken();
+						int colonIdx = token.indexOf(':');
+						String simpleSelectorText = colonIdx == -1 ? token : token.substring(0, colonIdx);
+						String pseudoElement = colonIdx == -1 ? null : token.substring(colonIdx+1);					
+						simpleSelectors.add(new SimpleSelector(simpleSelectorText, pseudoElement));
+						if(!tok.hasMoreTokens()) {
+							lastSelectorText = simpleSelectorText;
+							break;
 						}
-						ancestorSelectors.add(token);
-					}
-					else {
-						selector = token;
-						break;
 					}
 				}
-				if(selector != null) {
-					int colonIdx = selector.indexOf(':');
-					String pseudoName = colonIdx == -1 ? null : selector.substring(colonIdx+1);
-					String actualSelector = colonIdx == -1 ? selector : selector.substring(0, colonIdx);
-					int dotIdx = actualSelector.indexOf('.');
+				if(lastSelectorText != null) {
+					int dotIdx = lastSelectorText.indexOf('.');
 					if(dotIdx != -1) {
-						String elemtl = actualSelector.substring(0, dotIdx);
-						String classtl = actualSelector.substring(dotIdx+1);
-						this.addClassRule(elemtl, classtl, sr, ancestorSelectors, pseudoName);
+						String elemtl = lastSelectorText.substring(0, dotIdx);
+						String classtl = lastSelectorText.substring(dotIdx+1);
+						this.addClassRule(elemtl, classtl, sr, simpleSelectors);
 					}
 					else {
-						int poundIdx = actualSelector.indexOf('#');
+						int poundIdx = lastSelectorText.indexOf('#');
 						if(poundIdx != -1) {
-							String elemtl = actualSelector.substring(0, poundIdx);
-							String idtl = actualSelector.substring(poundIdx+1);
-							this.addIdRule(elemtl, idtl, sr, ancestorSelectors, pseudoName);
+							String elemtl = lastSelectorText.substring(0, poundIdx);
+							String idtl = lastSelectorText.substring(poundIdx+1);
+							this.addIdRule(elemtl, idtl, sr, simpleSelectors);
 						}
 						else {
-							String elemtl = actualSelector;
-							this.addElementRule(elemtl, sr, ancestorSelectors, pseudoName);
+							String elemtl = lastSelectorText;
+							this.addElementRule(elemtl, sr, simpleSelectors);
 						}
 					}
 				}
@@ -125,7 +121,7 @@ public class StyleSheetAggregator {
 				String href = importRule.getHref();
 				String styleHref = styleSheet.getHref();
 				String baseHref = styleHref == null ? document.getBaseURI() : styleHref;
-				CSSStyleSheet sheet = CSSUtilities.parse(href, document, baseHref, false);
+				CSSStyleSheet sheet = CSSUtilities.parse(styleSheet.getOwnerNode(), href, document, baseHref, false);
 				if(sheet != null) {
 					this.addStyleSheet(sheet);
 				}
@@ -145,7 +141,7 @@ public class StyleSheetAggregator {
 		}		
 	}
 
-	private final void addClassRule(String elemtl, String classtl, CSSStyleRule styleRule, ArrayList ancestorSelectors, String pseudoName) {
+	private final void addClassRule(String elemtl, String classtl, CSSStyleRule styleRule, ArrayList ancestorSelectors) {
 		Map classMap = (Map) this.classMapsByElement.get(elemtl);
 		if(classMap == null) {
 			classMap = new HashMap();
@@ -156,10 +152,10 @@ public class StyleSheetAggregator {
 			rules = new LinkedList();
 			classMap.put(classtl, rules);
 		}
-		rules.add(new StyleRuleInfo(ancestorSelectors, styleRule, pseudoName));
+		rules.add(new StyleRuleInfo(ancestorSelectors, styleRule));
 	}
 	
-	private final void addIdRule(String elemtl, String idtl, CSSStyleRule styleRule, ArrayList ancestorSelectors, String pseudoName) {
+	private final void addIdRule(String elemtl, String idtl, CSSStyleRule styleRule, ArrayList ancestorSelectors) {
 		Map idsMap = (Map) this.idMapsByElement.get(elemtl);
 		if(idsMap == null) {
 			idsMap = new HashMap();
@@ -170,16 +166,16 @@ public class StyleSheetAggregator {
 			rules = new LinkedList();
 			idsMap.put(idtl, rules);
 		}
-		rules.add(new StyleRuleInfo(ancestorSelectors, styleRule, pseudoName));
+		rules.add(new StyleRuleInfo(ancestorSelectors, styleRule));
 	}
 	
-	private final void addElementRule(String elemtl, CSSStyleRule styleRule, ArrayList ancestorSelectors, String pseudoName) {
+	private final void addElementRule(String elemtl, CSSStyleRule styleRule, ArrayList ancestorSelectors) {
 		Collection rules = (Collection) this.rulesByElement.get(elemtl);
 		if(rules == null) {
 			rules = new LinkedList();
 			this.rulesByElement.put(elemtl, rules);
 		}
-		rules.add(new StyleRuleInfo(ancestorSelectors, styleRule, pseudoName));
+		rules.add(new StyleRuleInfo(ancestorSelectors, styleRule));
 	}
 		
 	public final Collection getStyleDeclarations(HTMLElementImpl element, String elementName, String elementId, String className, Set pseudoNames) {
@@ -190,7 +186,7 @@ public class StyleSheetAggregator {
 			Iterator i = elementRules.iterator();
 			while(i.hasNext()) {
 				StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-				if(styleRuleInfo.matches(element, pseudoNames)) {
+				if(styleRuleInfo.isSelectorMatch(element, pseudoNames)) {
 					CSSStyleRule styleRule = styleRuleInfo.styleRule;
 					if(styleDeclarations == null) {
 						styleDeclarations = new LinkedList();
@@ -206,7 +202,7 @@ public class StyleSheetAggregator {
 			Iterator i = elementRules.iterator();
 			while(i.hasNext()) {
 				StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-				if(styleRuleInfo.matches(element, pseudoNames)) {
+				if(styleRuleInfo.isSelectorMatch(element, pseudoNames)) {
 					CSSStyleRule styleRule = styleRuleInfo.styleRule;
 					if(styleDeclarations == null) {
 						styleDeclarations = new LinkedList();
@@ -224,7 +220,7 @@ public class StyleSheetAggregator {
 					Iterator i = classRules.iterator();
 					while(i.hasNext()) {
 						StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-						if(styleRuleInfo.matches(element, pseudoNames)) {
+						if(styleRuleInfo.isSelectorMatch(element, pseudoNames)) {
 							CSSStyleRule styleRule = styleRuleInfo.styleRule;
 							if(styleDeclarations == null) {
 								styleDeclarations = new LinkedList();
@@ -241,7 +237,7 @@ public class StyleSheetAggregator {
 					Iterator i = classRules.iterator();
 					while(i.hasNext()) {
 						StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-						if(styleRuleInfo.matches(element, pseudoNames)) {
+						if(styleRuleInfo.isSelectorMatch(element, pseudoNames)) {
 							CSSStyleRule styleRule = styleRuleInfo.styleRule;
 							if(styleDeclarations == null) {
 								styleDeclarations = new LinkedList();
@@ -261,7 +257,7 @@ public class StyleSheetAggregator {
 					Iterator i = idRules.iterator();
 					while(i.hasNext()) {
 						StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-						if(styleRuleInfo.matches(element, pseudoNames)) {
+						if(styleRuleInfo.isSelectorMatch(element, pseudoNames)) {
 							CSSStyleRule styleRule = styleRuleInfo.styleRule;
 							if(styleDeclarations == null) {
 								styleDeclarations = new LinkedList();
@@ -279,7 +275,7 @@ public class StyleSheetAggregator {
 					Iterator i = idRules.iterator();
 					while(i.hasNext()) {
 						StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-						if(styleRuleInfo.matches(element, pseudoNames)) {
+						if(styleRuleInfo.isSelectorMatch(element, pseudoNames)) {
 							CSSStyleRule styleRule = styleRuleInfo.styleRule;
 							if(styleDeclarations == null) {
 								styleDeclarations = new LinkedList();
@@ -293,14 +289,14 @@ public class StyleSheetAggregator {
 		return styleDeclarations;
 	}
 
-	public final boolean hasPseudoName(HTMLElementImpl element, String elementName, String elementId, String[] classArray, String pseudoName) {
+	public final boolean affectedByPseudoNameInAncestor(HTMLElementImpl element, HTMLElementImpl ancestor, String elementName, String elementId, String[] classArray, String pseudoName) {
 		String elementTL = elementName.toLowerCase();
 		Collection elementRules = (Collection) this.rulesByElement.get(elementTL);
 		if(elementRules != null) {
 			Iterator i = elementRules.iterator();
 			while(i.hasNext()) {
 				StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-				if(styleRuleInfo.hasPseudoNameMatch(element, pseudoName)) {
+				if(styleRuleInfo.affectedByPseudoNameInAncestor(element, ancestor, pseudoName)) {
 					return true;
 				}
 			}
@@ -310,7 +306,7 @@ public class StyleSheetAggregator {
 			Iterator i = elementRules.iterator();
 			while(i.hasNext()) {
 				StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-				if(styleRuleInfo.hasPseudoNameMatch(element, pseudoName)) {
+				if(styleRuleInfo.affectedByPseudoNameInAncestor(element, ancestor, pseudoName)) {
 					return true;
 				}
 			}
@@ -326,7 +322,7 @@ public class StyleSheetAggregator {
 						Iterator i = classRules.iterator();
 						while(i.hasNext()) {
 							StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-							if(styleRuleInfo.hasPseudoNameMatch(element, pseudoName)) {
+							if(styleRuleInfo.affectedByPseudoNameInAncestor(element, ancestor, pseudoName)) {
 								return true;
 							}
 						}
@@ -339,7 +335,7 @@ public class StyleSheetAggregator {
 						Iterator i = classRules.iterator();
 						while(i.hasNext()) {
 							StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-							if(styleRuleInfo.hasPseudoNameMatch(element, pseudoName)) {
+							if(styleRuleInfo.affectedByPseudoNameInAncestor(element, ancestor, pseudoName)) {
 								return true;
 							}
 						}
@@ -356,7 +352,7 @@ public class StyleSheetAggregator {
 					Iterator i = idRules.iterator();
 					while(i.hasNext()) {
 						StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-						if(styleRuleInfo.hasPseudoNameMatch(element, pseudoName)) {
+						if(styleRuleInfo.affectedByPseudoNameInAncestor(element, ancestor, pseudoName)) {
 							return true;
 						}
 					}				
@@ -370,7 +366,7 @@ public class StyleSheetAggregator {
 					Iterator i = idRules.iterator();
 					while(i.hasNext()) {
 						StyleRuleInfo styleRuleInfo = (StyleRuleInfo) i.next();
-						if(styleRuleInfo.hasPseudoNameMatch(element, pseudoName)) {
+						if(styleRuleInfo.affectedByPseudoNameInAncestor(element, ancestor, pseudoName)) {
 							return true;
 						}
 					}				
@@ -383,76 +379,161 @@ public class StyleSheetAggregator {
 	private static class StyleRuleInfo {
 		private final CSSStyleRule styleRule;
 		private final ArrayList ancestorSelectors;
-		private final String pseudoName;
 
 		/**
-		 * @param selectors A collection of selectors already in lower case.
+		 * @param selectors A collection of SimpleSelector's.
 		 * @param rule A CSS rule.
 		 */
-		public StyleRuleInfo(ArrayList selectors, CSSStyleRule rule, String pseudoName) {
+		public StyleRuleInfo(ArrayList simpleSelectors, CSSStyleRule rule) {
 			super();
-			ancestorSelectors = selectors;
+			ancestorSelectors = simpleSelectors;
 			styleRule = rule;
-			this.pseudoName = pseudoName;
 		}
 
-		public boolean hasPseudoNameMatch(HTMLElementImpl element, String pseudoName) {
-			String localPn = this.pseudoName;
-			if(localPn == null || !pseudoName.equalsIgnoreCase(localPn)) {
-				return false;
-			}
-			return this.isSelectorMatch(element);
-		}
-		
-		public boolean matches(HTMLElementImpl element, Set pseudoNames) {
-			if(this.pseudoName != null && (pseudoNames == null || !pseudoNames.contains(this.pseudoName))) {
-				return false;
-			}
-			return this.isSelectorMatch(element);
-		}
-		
-		private boolean isSelectorMatch(HTMLElementImpl element) {
+		public final boolean affectedByPseudoNameInAncestor(HTMLElementImpl element, HTMLElementImpl ancestor, String pseudoName) {
 			ArrayList as = this.ancestorSelectors;
-			if(as == null) {
-				return true;
-			}
 			HTMLElementImpl currentElement = element;
 			int size = as.size();
+			boolean first = true;
 			for(int i = size; --i >= 0;) {
-				String selector = (String) as.get(i);
-				int dotIdx = selector.indexOf('.');
-				if(dotIdx != -1) {
-					String elemtl = selector.substring(0, dotIdx);
-					String classtl = selector.substring(dotIdx+1);
-					HTMLElementImpl ancestor = currentElement.getAncestorWithClass(elemtl, classtl);
-					if(ancestor == null) {
-						return false;
+				SimpleSelector simpleSelector = (SimpleSelector) as.get(i);
+				if(first) {
+					if(ancestor == element) {
+						return simpleSelector.hasPseudoName(pseudoName);
 					}
-					currentElement = ancestor;					
+					first = false;
+					continue;
+				}
+				String selectorText = simpleSelector.simpleSelectorText;
+				int dotIdx = selectorText.indexOf('.');
+				HTMLElementImpl newElement;
+				if(dotIdx != -1) {
+					String elemtl = selectorText.substring(0, dotIdx);
+					String classtl = selectorText.substring(dotIdx+1);
+					newElement = currentElement.getAncestorWithClass(elemtl, classtl);
 				}
 				else {
-					int poundIdx = selector.indexOf('#');
+					int poundIdx = selectorText.indexOf('#');
 					if(poundIdx != -1) {
-						String elemtl = selector.substring(0, poundIdx);
-						String idtl = selector.substring(poundIdx+1);
-						HTMLElementImpl ancestor = currentElement.getAncestorWithId(elemtl, idtl);
-						if(ancestor == null) {
-							return false;
-						}
-						currentElement = ancestor;											
+						String elemtl = selectorText.substring(0, poundIdx);
+						String idtl = selectorText.substring(poundIdx+1);
+						newElement = currentElement.getAncestorWithId(elemtl, idtl);
 					}
 					else {
-						String elemtl = selector;
-						HTMLElementImpl ancestor = currentElement.getAncestor(elemtl);
-						if(ancestor == null) {
-							return false;
-						}
-						currentElement = ancestor;											
+						String elemtl = selectorText;
+						newElement = currentElement.getAncestor(elemtl);
 					}
 				}				
+				if(newElement == null) {
+					return false;
+				}
+				currentElement = newElement;					
+				if(currentElement == ancestor) {
+					return simpleSelector.hasPseudoName(pseudoName);
+				}
+			}
+			return false;			
+		}
+		
+		/**
+		 * 
+		 * @param element The element to test for a match.
+		 * @param pseudoNames A set of pseudo-names in lowercase.
+		 */
+		private final boolean isSelectorMatch(HTMLElementImpl element, Set pseudoNames) {
+			ArrayList as = this.ancestorSelectors;
+			HTMLElementImpl currentElement = element;
+			int size = as.size();
+			boolean first = true;
+			for(int i = size; --i >= 0;) {
+				SimpleSelector simpleSelector = (SimpleSelector) as.get(i);
+				if(first) {
+					if(!simpleSelector.matches(pseudoNames)) {
+						return false;
+					}
+					first = false;
+					continue;
+				}
+				String selectorText = simpleSelector.simpleSelectorText;
+				int dotIdx = selectorText.indexOf('.');
+				HTMLElementImpl ancestor;
+				if(dotIdx != -1) {
+					String elemtl = selectorText.substring(0, dotIdx);
+					String classtl = selectorText.substring(dotIdx+1);
+					ancestor = currentElement.getAncestorWithClass(elemtl, classtl);
+				}
+				else {
+					int poundIdx = selectorText.indexOf('#');
+					if(poundIdx != -1) {
+						String elemtl = selectorText.substring(0, poundIdx);
+						String idtl = selectorText.substring(poundIdx+1);
+						ancestor = currentElement.getAncestorWithId(elemtl, idtl);
+					}
+					else {
+						String elemtl = selectorText;
+						ancestor = currentElement.getAncestor(elemtl);
+					}
+				}				
+				if(ancestor == null) {
+					return false;
+				}
+				if(!simpleSelector.matches(ancestor)) {
+					return false;
+				}
+				currentElement = ancestor;					
 			}
 			return true;
 		}
 	}
 	
+	private static class SimpleSelector {
+		public String simpleSelectorText;
+		public String pseudoElement;
+
+		/**
+		 * 
+		 * @param simpleSelectorText Simple selector text in lower case.
+		 * @param pseudoElement The pseudo-element if any.
+		 */
+		public SimpleSelector(String simpleSelectorText, String pseudoElement) {
+			super();
+			this.simpleSelectorText = simpleSelectorText;
+			this.pseudoElement = pseudoElement;
+		}
+		
+		public final boolean matches(HTMLElementImpl element) {
+			Set names = element.getPseudoNames();
+			if(names == null) {
+				return this.pseudoElement == null;
+			}
+			else {
+				String pe = this.pseudoElement;
+				return pe == null || names.contains(pe);
+			}
+		}
+
+		public final boolean matches(Set names) {
+			if(names == null) {
+				return this.pseudoElement == null;
+			}
+			else {
+				String pe = this.pseudoElement;
+				return pe == null || names.contains(pe);
+			}
+		}
+		
+		public final boolean matches(String pseudoName) {
+			if(pseudoName == null) {
+				return this.pseudoElement == null;
+			}
+			else {
+				String pe = this.pseudoElement;
+				return pe == null || pseudoName.equals(pe);
+			}
+		}
+		
+		public final boolean hasPseudoName(String pseudoName) {
+			return pseudoName.equals(this.pseudoElement);
+		}
+	}
 }
