@@ -16,7 +16,7 @@
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    Contact info: xamjadmin@users.sourceforge.net
+    Contact info: lobochief@users.sourceforge.net
 */
 /*
  * Created on Apr 16, 2005
@@ -139,7 +139,7 @@ public class StyleSheetRenderState implements RenderState {
 		// Dummy implementation
 	}
 
-	protected final CSS2PropertiesImpl getCssProperties() {
+	protected final AbstractCSS2Properties getCssProperties() {
 		HTMLElementImpl element = this.element;
 		return element == null ? null : element.getCurrentStyle();
 	}
@@ -221,7 +221,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(f != null) {
 			return f;
 		}
-		CSS2PropertiesImpl style = this.getCssProperties();
+		AbstractCSS2Properties style = this.getCssProperties();
 		RenderState prs = this.prevRenderState;
 		if(style == null) {
 			if(prs != null) {
@@ -239,21 +239,12 @@ public class StyleSheetRenderState implements RenderState {
 		String fontWeight = null;
 		String fontFamily = null;
 		
-		String fontSpec = style == null ? null : style.getFont();
-		if(fontSpec != null) {
-			FontInfo fontInfo = HtmlValues.createFontInfo(fontSpec, this.prevRenderState);
-			fontSize = fontInfo.fontSize;
-			fontStyle = fontInfo.fontStyle;
-			fontVariant = fontInfo.fontVariant;
-			fontWeight = fontInfo.fontWeight;
-			fontFamily = fontInfo.fontFamily;
-		}		
 		String newFontSize = style == null ? null : style.getFontSize();
 		String newFontFamily = style == null ? null : style.getFontFamily();
 		String newFontStyle = style == null ? null : style.getFontStyle();
 		String newFontVariant = style == null ? null : style.getFontVariant();
 		String newFontWeight = style == null ? null : style.getFontWeight();
-		if(fontSpec == null && newFontSize == null && newFontWeight == null && newFontStyle == null && newFontFamily == null && newFontVariant == null) {
+		if(newFontSize == null && newFontWeight == null && newFontStyle == null && newFontFamily == null && newFontVariant == null) {
 			if(prs != null) {
 				f = prs.getFont();
 				this.iFont = f;
@@ -320,7 +311,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(c != null) {
 			return c;
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		String colorValue = props == null ? null : props.getColor();
 		if(colorValue == null || "".equals(colorValue)) {
 			RenderState prs = this.prevRenderState;
@@ -393,7 +384,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(c != INVALID_COLOR) {
 			return c;
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		String colorValue = props == null ? null : props.getOverlayColor();
 		if(colorValue == null || colorValue.length() == 0) {
 			RenderState prs = this.prevRenderState;
@@ -416,7 +407,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(td != -1) {
 			return td;
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		String tdText = props == null ? null : props.getTextDecoration();
 		if(tdText == null) {
 			RenderState prs = this.prevRenderState;
@@ -623,16 +614,8 @@ public class StyleSheetRenderState implements RenderState {
 			return binfo;
 		}
 		binfo = null;
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		if(props != null) {
-			String background = props.getBackground();
-			if(background != null) {
-				if(binfo == null) {
-					binfo = new BackgroundInfo();
-				}
-				CSSStyleDeclaration backgroundDecl = props.getStyleDeclaration("background");
-				this.applyBackground(binfo, background, backgroundDecl);
-			}
 			String backgroundColorText = props.getBackgroundColor();
 			if(backgroundColorText != null) {
 				if(binfo == null) {
@@ -641,9 +624,8 @@ public class StyleSheetRenderState implements RenderState {
 				binfo.backgroundColor = ColorFactory.getInstance().getColor(backgroundColorText);
 			}		
 			String backgroundImageText = props.getBackgroundImage();
-			if(backgroundImageText != null) {
-				CSSStyleDeclaration backgroundImageDecl = props.getStyleDeclaration("background-image");
-				java.net.URL backgroundImage = HtmlValues.getURIFromStyleValue(backgroundImageText, backgroundImageDecl, this.document);
+			if(backgroundImageText != null && backgroundImageText.length() > 0) {
+				java.net.URL backgroundImage = HtmlValues.getURIFromStyleValue(backgroundImageText);
 				if(backgroundImage != null) {
 					if(binfo == null) {
 						binfo = new BackgroundInfo();
@@ -677,7 +659,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(tiText != null) {
 			return tiText;
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		tiText = props == null ? null : props.getTextIndent();
 		if(tiText == null) {
 			RenderState prs = this.prevRenderState;
@@ -714,7 +696,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(ws != null) {
 			return ws.intValue();
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		String whiteSpaceText = props == null ? null : props.getWhiteSpace();
 		int wsValue;
 		if(whiteSpaceText == null) {
@@ -756,7 +738,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(mi != INVALID_INSETS) {
 			return mi;
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		if(props == null) { 
 			mi = null;
 		}
@@ -772,7 +754,7 @@ public class StyleSheetRenderState implements RenderState {
 		if(mi != INVALID_INSETS) {
 			return mi;
 		}
-		CSS2PropertiesImpl props = this.getCssProperties();
+		AbstractCSS2Properties props = this.getCssProperties();
 		if(props == null) {
 			mi = null;
 		}
@@ -849,46 +831,31 @@ public class StyleSheetRenderState implements RenderState {
 		}
 	}
 	
-	private void applyBackground(BackgroundInfo binfo, String background, CSSStyleDeclaration declaration) {
-		String[] tokens = HtmlValues.splitCssValue(background);
-		boolean hasXPosition = false;
-		for(int i = 0; i < tokens.length; i++) {
-			String token = tokens[i];
-			if(ColorFactory.getInstance().isColor(token)) {
-				binfo.backgroundColor = ColorFactory.getInstance().getColor(token);
-			}	
-			else if(HtmlValues.isUrl(token)) {
-				binfo.backgroundImage = HtmlValues.getURIFromStyleValue(token, declaration, this.document);
-			}
-			else if(isBackgroundRepeat(token)) {
-				this.applyBackgroundRepeat(binfo, token);
-			}
-			else if(isBackgroundPosition(token)) {
-				if(hasXPosition) {
-					this.applyBackgroundVerticalPosition(binfo, token);
-				}
-				else {
-					hasXPosition = true;
-					this.applyBackgroundHorizontalPositon(binfo, token);
-				}
-			}
-		}
-	}
-
-	private boolean isBackgroundPosition(String token) {
-		return HtmlValues.isLength(token) ||
-			token.endsWith("%") ||
-			token.equalsIgnoreCase("top") ||
-			token.equalsIgnoreCase("center") ||
-			token.equalsIgnoreCase("bottom") ||
-			token.equalsIgnoreCase("left") ||
-			token.equalsIgnoreCase("right");			
-	}
-	
-	private boolean isBackgroundRepeat(String repeat) {
-		String repeatTL = repeat.toLowerCase();
-		return repeatTL.indexOf("repeat") != -1;
-	}
+//	private void applyBackground(BackgroundInfo binfo, String background, CSSStyleDeclaration declaration) {
+//		String[] tokens = HtmlValues.splitCssValue(background);
+//		boolean hasXPosition = false;
+//		for(int i = 0; i < tokens.length; i++) {
+//			String token = tokens[i];
+//			if(ColorFactory.getInstance().isColor(token)) {
+//				binfo.backgroundColor = ColorFactory.getInstance().getColor(token);
+//			}	
+//			else if(HtmlValues.isUrl(token)) {
+//				binfo.backgroundImage = HtmlValues.getURIFromStyleValue(token, declaration, this.document);
+//			}
+//			else if(isBackgroundRepeat(token)) {
+//				this.applyBackgroundRepeat(binfo, token);
+//			}
+//			else if(isBackgroundPosition(token)) {
+//				if(hasXPosition) {
+//					this.applyBackgroundVerticalPosition(binfo, token);
+//				}
+//				else {
+//					hasXPosition = true;
+//					this.applyBackgroundHorizontalPositon(binfo, token);
+//				}
+//			}
+//		}
+//	}
 
 	private void applyBackgroundRepeat(BackgroundInfo binfo, String backgroundRepeatText) {
 		String brtl = backgroundRepeatText.toLowerCase();
@@ -906,7 +873,117 @@ public class StyleSheetRenderState implements RenderState {
 		}		
 	}
 
+	private Integer cachedVisibility;
 	
+	public int getVisibility() {
+		Integer v = this.cachedVisibility;
+		if(v != null) {
+			return v.intValue();
+		}
+		AbstractCSS2Properties props = this.getCssProperties();
+		int visibility;
+		if(props == null) {
+			visibility = VISIBILITY_VISIBLE;
+		}
+		else {
+			String visibText = props.getVisibility();
+			if(visibText == null || visibText.length() == 0) {
+				visibility = VISIBILITY_VISIBLE;
+			}
+			else {
+				String visibTextTL = visibText.toLowerCase();
+				if(visibTextTL.equals("hidden")) {
+					visibility = VISIBILITY_HIDDEN;
+				}
+				else if(visibTextTL.equals("visible")) {
+					visibility = VISIBILITY_VISIBLE;
+				}
+				else if(visibTextTL.equals("collapse")) {
+					visibility = VISIBILITY_COLLAPSE;
+				}
+				else {
+					visibility = VISIBILITY_VISIBLE;
+				}
+			}
+		}
+		this.cachedVisibility = new Integer(visibility);
+		return visibility;
+	}
+	
+	private Integer cachedPosition;
+	
+	public int getPosition() {
+		Integer p = this.cachedPosition;
+		if(p != null) {
+			return p.intValue();
+		}
+		AbstractCSS2Properties props = this.getCssProperties();
+		int position;
+		if(props == null) {
+			position = POSITION_STATIC;
+		}
+		else {
+			String positionText = props.getPosition();
+			if(positionText == null || positionText.length() == 0) {
+				position = POSITION_STATIC;
+			}
+			else {
+				String positionTextTL = positionText.toLowerCase();
+				if(positionTextTL.equals("absolute")) {
+					position = POSITION_ABSOLUTE;
+				}
+				else if(positionTextTL.equals("static")) {
+					position = POSITION_STATIC;
+				}
+				else if(positionTextTL.equals("relative")) {
+					position = POSITION_RELATIVE;
+				}
+				else if(positionTextTL.equals("fixed")) {
+					position = POSITION_FIXED;
+				}
+				else {
+					position = POSITION_STATIC;
+				}
+			}
+		}
+		this.cachedPosition = new Integer(position);
+		return position;		
+	}
+
+	private Integer cachedFloat;
+	
+	public int getFloat() {
+		Integer p = this.cachedFloat;
+		if(p != null) {
+			return p.intValue();
+		}
+		AbstractCSS2Properties props = this.getCssProperties();
+		int floatValue;
+		if(props == null) {
+			floatValue = FLOAT_NONE;
+		}
+		else {
+			String floatText = props.getFloat();
+			if(floatText == null || floatText.length() == 0) {
+				floatValue = FLOAT_NONE;
+			}
+			else {
+				String floatTextTL = floatText.toLowerCase();
+				if(floatTextTL.equals("left")) {
+					floatValue = FLOAT_LEFT;
+				}
+				else if(floatTextTL.equals("right")) {
+					floatValue = FLOAT_RIGHT;
+				}
+				else {
+					floatValue = FLOAT_NONE;
+				}
+			}
+		}
+		this.cachedFloat = new Integer(floatValue);
+		return floatValue;		
+	}
+
 	public String toString() {
 		return "StyleSheetRenderState[font=" + this.getFont() + ",textDecoration=" + this.getTextDecorationMask() + "]";
 	}
